function output_member = filter_ic_cbmember_smc(filter, model, meas, truth)
% output_member = filter_ic_cbmember_smc(model, meas, truth)
% INPUT:
%      model - structure that contains kinematic and observation params
%      meas  - structure that contains measurements
%      truth - structure that contains ground truth tracks
% OUTPUT:
%      output_member - structure that hold output estimates
% -------------------------------------------------------------------------
% Function that implements an Iterated Corector Cardinality Balanced 
% multi-Bernoulli (IC-CBMeMBer) filter. The CBMeMBer filter was proposed in
% [1]. The IC-CBMeMBer sequentially applies the update stap of the CBMeMBer 
% for each sensor. The state models are non-linear and the implementation
% is based on the particle filter (SMC). 
% -------------------------------------------------------------------------
% [1] B.-T. Vo, B.-N. Vo, and A. Cantoni, "The cardinality balanced multitarget
% multi-Bernoulli filter and its implementations," IEEE Trans. Signal
% Process., vol. 57, no. 2, pp. 409 - 423, Feb. 2009.
%
% Contact: augustin.saucan@mail.mcgill.ca

%=== Setup
%output variables
output_member.state_X_hat = cell(meas.K,1);   % inferred target states
output_member.state_n_hat = zeros(meas.K,1);  % estimated number of targets
output_member.filter      = filter;           % store filtering params

%=== Filtering
member_mark_start = tic;      % start timing the filter

% void initial prior
SCB_MeMBer_update.M = 0;       % nb of Bernoulli components
SCB_MeMBer_update.r = [];      % prob of existence of component
SCB_MeMBer_update.x = {};      % particles of component
SCB_MeMBer_update.w = {};      % wights of particles

%recursive filtering
for k = 1:meas.K
    
    % Prediction and update
    SCB_MeMBer_predict = predict(SCB_MeMBer_update, model, filter);                                     H_predict = SCB_MeMBer_predict.M;
    SCB_MeMBer_update  = ms_update(SCB_MeMBer_predict, model, filter, meas.Z{k});                 H_posterior = SCB_MeMBer_update.M;
    
    % Pruning 
    SCB_MeMBer_update = prune(SCB_MeMBer_update, filter);                                               H_prune = SCB_MeMBer_update.M;

    % State estimation
    [output_member.state_X_hat{k}, output_member.state_n_hat(k)] = extract_estimates(SCB_MeMBer_update);
    
    % Cap the Bernoulli components at either the maximum number of allowed
    % components or to a number proportional with the number of estimated
    % targets
    N_cap = min([filter.comp_max, SCB_MeMBer_update.M, max(output_member.state_n_hat(k) * filter.nb_comp_tg, filter.nb_comp_tg) ]);
    SCB_MeMBer_update = cap(SCB_MeMBer_update, N_cap);                                                   H_cap = SCB_MeMBer_update.M;
    
    % Return the updated Bernoulli existence probabilities for diagnostics
    output_member.rkk{k} = SCB_MeMBer_update.r;  
    
    % Display diagnostics
    display_diaginfo( k, output_member, filter, H_predict, H_posterior, H_prune, H_cap);  
end

% return filter computation time 
output_member.all_time    = toc(member_mark_start);

% calculate and display average OSPA error
error_t = zeros(1,truth.K);
for k = 1:truth.K
    error_t(k) = ospa_dist( truth.X{k}, output_member.state_X_hat{k}, model.ospa_cutoff, model.ospa_order);
end

% return OSPA errors
output_member.ospa = error_t;

end % end function display

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
%% %%%%%%%%%%%%%%%%%%%%%%%%%%% LOCAL FUNCTIONS %%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function SCB_MeMBer_predict = predict(SCB_MeMBer_update, model, filter)

% Predict surviving tracks
SCB_MeMBer_predict   = SCB_MeMBer_update;
SCB_MeMBer_predict.r = SCB_MeMBer_update.r * model.kin.pS;
for ii = 1:SCB_MeMBer_update.M
    % predict particles of each Bernoulli component 
    SCB_MeMBer_predict.x{ii} = gen_newstate_fn(model, SCB_MeMBer_update.x{ii}, 'noise');  
end

% append birth Bernoulli components
for ii = 1:model.birth.J_gamma
    SCB_MeMBer_predict.r(SCB_MeMBer_predict.M + ii) = model.birth.r_gamma(ii);
    SCB_MeMBer_predict.x{SCB_MeMBer_predict.M + ii} = repmat(model.birth.m_gamma(:,ii), 1, filter.nb_part) + model.birth.sqrt_P_gamma * randn(model.kin.x_dim, filter.nb_part);
    SCB_MeMBer_predict.w{SCB_MeMBer_predict.M + ii} = 1/filter.nb_part * ones(1,filter.nb_part);
end

% account for the increase in the Bernoulli components
SCB_MeMBer_predict.M = SCB_MeMBer_update.M + model.birth.J_gamma;

% clipping of existance probabilities
SCB_MeMBer_predict.r = limit_range(SCB_MeMBer_predict.r);

end

function SCB_MeMBer_sens_up = ms_update(SCB_MeMBer_predict, model, filter, Z)

SCB_MeMBer_sens_up = SCB_MeMBer_predict;

for s_idx = 1:model.obs.nb_sensors
    SCB_MeMBer_sens_up = ss_update(SCB_MeMBer_sens_up, model, filter, Z{s_idx}, s_idx);
end
    
end

function SCB_MeMBer_update = ss_update(SCB_MeMBer_predict, model, filter, Z, s_idx)

    %---construction of pseudo PHD for update
    J_pseudo  = SCB_MeMBer_predict.M * filter.nb_part;                     %number of particles in pseudo-PHD
    x_pseudo  = zeros(model.kin.x_dim, J_pseudo);                          %states of particles in pseudo-PHD
    w_pseudo  = zeros(1, J_pseudo);                                        %weights of particles in pseudo-PHD
    w_pseudo1 = zeros(1, J_pseudo);                                        %alt weight (1) of particles in pseudo-PHD - used in CB-MeMBer update later
    w_pseudo2 = zeros(1, J_pseudo);                                        %alt weight (2) of particles in pseudo-PHD - used in CB-MeMBer update later
    

    for t = 1:SCB_MeMBer_predict.M                  
        idx_prt = [1:filter.nb_part] + (t-1)*filter.nb_part;
        
        w_temp = model.obs.compute_pD(SCB_MeMBer_predict.x{t}, s_idx)' .* SCB_MeMBer_predict.w{t};
        x_pseudo(:,idx_prt) = SCB_MeMBer_predict.x{t};
        
        w_pseudo(idx_prt)   = SCB_MeMBer_predict.r(t)/(1-SCB_MeMBer_predict.r(t)) * SCB_MeMBer_predict.w{t};
        w_pseudo1(idx_prt)  = SCB_MeMBer_predict.r(t)/(1-SCB_MeMBer_predict.r(t) * sum(w_temp)) * SCB_MeMBer_predict.w{t};
        w_pseudo2(idx_prt)  = SCB_MeMBer_predict.r(t)*(1-SCB_MeMBer_predict.r(t))/((1-SCB_MeMBer_predict.r(t) * sum(w_temp))^2) * SCB_MeMBer_predict.w{t};        
       
    end

    % number of measurements
    m = size(Z,2);
    
    SCB_MeMBer_update.M = SCB_MeMBer_predict.M + m;                                                                   %total number of tracks/components
    SCB_MeMBer_update.r = zeros(SCB_MeMBer_update.M,1);                                                                                  %existence probability for tracks/components
    SCB_MeMBer_update.x = cell(1,SCB_MeMBer_update.M);                                                                                   %states of particles in each track/component
    SCB_MeMBer_update.w = cell(1,SCB_MeMBer_update.M);                                                                                   %weights of particles in each track/component
 
    %legacy tracks
    for t = 1:SCB_MeMBer_predict.M  
        w_temp = SCB_MeMBer_predict.w{t} .* model.obs.compute_pD(SCB_MeMBer_predict.x{t}, s_idx)';

        SCB_MeMBer_update.r(t) = SCB_MeMBer_predict.r(t)*((1-sum(w_temp))/(1-SCB_MeMBer_predict.r(t)*sum(w_temp)));
        SCB_MeMBer_update.x{t} = SCB_MeMBer_predict.x{t};
        SCB_MeMBer_update.w{t} = SCB_MeMBer_predict.w{t}.*(1-model.obs.compute_pD(SCB_MeMBer_predict.x{t}, s_idx)')/(1-sum(w_temp));
        
        SCB_MeMBer_update.w{t} = SCB_MeMBer_update.w{t} / sum(SCB_MeMBer_update.w{t});
                
    end
    
    %measurement updated tracks
    if m ~= 0
        offset = SCB_MeMBer_predict.M;
        pD_vals = model.obs.compute_pD(x_pseudo, s_idx)';
        
        z_pseudo = model.obs.gen_sen_obs(s_idx, x_pseudo, []);
        
        for z_idx = 1:m           
            meas_likelihood = exp(model.obs.lik_fn(repmat(Z(:,z_idx),1,size(z_pseudo,2)), z_pseudo, model.obs.R));
            
            r_temp = sum(pD_vals.*w_pseudo2.*meas_likelihood) / (model.obs.lambda_c*model.obs.pdf_c+sum(pD_vals.*w_pseudo1.*meas_likelihood));
            w_temp = pD_vals.*w_pseudo.*meas_likelihood; 
            
            if (sum(w_temp) == 0) || (sum(isnan(w_temp)) > 0)
                w_temp = ones(1,length(w_temp)) / length(w_temp);
            else
                w_temp = w_temp/sum(w_temp);
            end
            
            [ ~, ~, rsp_idx ] = Resampling( x_pseudo, w_temp , filter.nb_part, 'stratified' ); % resample particles

            
            SCB_MeMBer_update.r(offset+z_idx) = r_temp;
            SCB_MeMBer_update.x{offset+z_idx} = x_pseudo(:, rsp_idx);
            SCB_MeMBer_update.w{offset+z_idx} = w_temp(rsp_idx) / sum(w_temp(rsp_idx));             
        end
        
    end

    SCB_MeMBer_update.r = limit_range(SCB_MeMBer_update.r);   

end % end function update


function member_out = prune(member_in, filter)
% prune Bernoulli components with weights below a given threshold
idx = find( member_in.r > filter.comp_threshold );

member_out.M = length(idx);
member_out.r = member_in.r(idx);

for ii = 1:length(idx)
    member_out.x{ii} = member_in.x{idx(ii)};
    member_out.w{ii} = member_in.w{idx(ii)};
end
end % end function prune


function member_out = cap(member_in, N_cap)
% Cap total number of Bernoulli components to a specified maximum

if member_in.M > N_cap
    [~, idxsort] = sort(member_in.r,'descend');
    idxkeep = idxsort(1:N_cap);

    member_out.M = length(idxkeep);
    member_out.r = member_in.r(idxkeep);
    for ii = 1:length(idxkeep)
        member_out.x{ii} = member_in.x{idxkeep(ii)};
        member_out.w{ii} = member_in.w{idxkeep(ii)};
    end   
else
    member_out = member_in;
end

end % end function cap


function [X, N] = extract_estimates(member)
% State inference from the updated multi Bernoulli distribution
cdn_pmf = prod(1-member.r) * esf(member.r ./ (1-member.r));

[~, idx_max_cdn] = max(cdn_pmf);
map_cdn = idx_max_cdn-1;
N = min(member.M, map_cdn);   % estimate number of targets

[~, idx_max_tgt] = sort(member.r, 'descend');
X = [];
for t = 1:N % compute sample means for each N estimated targets
    X(:,t) =  member.x{idx_max_tgt(t)} * member.w{idx_max_tgt(t)}(:);
end

end % end function extract estimates


function display_diaginfo( k, est, filter, H_predict, H_posterior, H_prune, H_cap)
% Display diagnostic information on the fly
if ~strcmp(filter.run_flag,'silence')
    disp([' time= ',num2str(k),...
          ' #est card=' num2str(est.state_n_hat(k),4),...
          ' #comp pred=' num2str(H_predict,4),...
          ' #comp post=' num2str(H_posterior,4),...
          ' #comp pruned=',num2str(H_prune),...
          ' #comp updt=',num2str(H_cap,4)   ]);
end
end % end function display