function output_member = filter_ic_cbmember_ukf(filter, model, meas, truth)
% output_member = filter_ic_cbmember_ukf(model, meas, truth)
% INPUT:
%      model - structure that contains kinematic and observation params
%      meas  - structure that contains measurements
%      truth - structure that contains ground truth tracks
% OUTPUT:
%      output_member - structure that hold output estimates
% -------------------------------------------------------------------------
% Function that implements an Iterated Corector Cardinality Balanced 
% multi-Bernoulli (IC-CBMeMBer) filter. The CBMeMBer filter was proposed in
% [1]. The IC-CBMeMBer sequentially applies the update stap of the CBMeMBer 
% for each sensor. The state models are non-linear, hence the 
% implementation is based on the UKF. 
% -------------------------------------------------------------------------
% [1] B.-T. Vo, B.-N. Vo, and A. Cantoni, "The cardinality balanced multitarget
% multi-Bernoulli filter and its implementations," IEEE Trans. Signal
% Process., vol. 57, no. 2, pp. 409 - 423, Feb. 2009.
%
% Contact: augustin.saucan@mail.mcgill.ca

%=== Setup

%output variables
output_member.state_X_hat = cell(meas.K,1);   % inferred target states
output_member.state_n_hat = zeros(meas.K,1);  % estimated number of targets
output_member.filter      = filter;           % store filtering params

%=== Filtering
member_mark_start = tic;      % start timing the filter

% Setup UKF params
filter.ukf.L      = model.kin.x_dim + model.obs.z_dim;  % number of sigma points is 2*filter.ukf.L +1
filter.ukf.Wm     = [filter.ukf.lambda/(filter.ukf.L+filter.ukf.lambda), (1/2/(filter.ukf.L+filter.ukf.lambda))*ones(1,2*filter.ukf.L)];   % Sigma point weights
filter.ukf.Wc     = filter.ukf.Wm;

% void initial prior
SCB_MeMBer_update.M = 0;       % nb of Bernoulli components
SCB_MeMBer_update.r = [];      % prob of existence of Bernoulli component
SCB_MeMBer_update.L = [];      % number of mixture components for each Bernoulli component
SCB_MeMBer_update.m = {};      % means of mixture components of each Bernoulli component
SCB_MeMBer_update.P = {};      % covariances of mixture components of each Bernoulli component
SCB_MeMBer_update.w = {};      % mixture weights of each Bernoulli component

%recursive filtering
for k = 1:meas.K
    
    % Prediction and update
    SCB_MeMBer_predict = predict(SCB_MeMBer_update, model);                                    H_predict = SCB_MeMBer_predict.M;
    SCB_MeMBer_update  = ms_update(SCB_MeMBer_predict, model, filter, meas.Z{k});                      H_posterior = SCB_MeMBer_update.M;
    
   
    % Pruning of Bernoulli components 
    SCB_MeMBer_update = prune(SCB_MeMBer_update, filter);                                                H_prune = SCB_MeMBer_update.M;

    % State estimation
    [output_member.state_X_hat{k}, output_member.state_n_hat(k)] = extract_estimates(SCB_MeMBer_update);
    
    % Cap the Bernoulli components at either the maximum number of allowed
    % components or to a number proportional with the number of estimated
    % targets
    N_cap = min([filter.comp_max, SCB_MeMBer_update.M, max(output_member.state_n_hat(k) * filter.nb_comp_tg, filter.nb_comp_tg) ]);
    SCB_MeMBer_update = cap(SCB_MeMBer_update, N_cap);                                                   H_cap = SCB_MeMBer_update.M;
    
    % Return the updated Bernoulli existence probabilities for diagnostics
    output_member.rkk{k} = SCB_MeMBer_update.r;  
    
    % Display diagnostics
    display_diaginfo( k, output_member, filter, H_predict, H_posterior, H_prune, H_cap);  
end

% return filter computation time 
output_member.all_time    = toc(member_mark_start);

% calculate and display average OSPA error
error_t = zeros(1,truth.K);
for k = 1:truth.K
    error_t(k) = ospa_dist( truth.X{k}, output_member.state_X_hat{k}, model.ospa_cutoff, model.ospa_order);
end

% return OSPA errors
output_member.ospa = error_t;

end % end function display

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
%% %%%%%%%%%%%%%%%%%%%%%%%%%%% LOCAL FUNCTIONS %%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function SCB_MeMBer_predict = predict(SCB_MeMBer_update, model)

% Predict surviving tracks
SCB_MeMBer_predict   = SCB_MeMBer_update;
SCB_MeMBer_predict.r = SCB_MeMBer_update.r * model.kin.pS;
for ii = 1:SCB_MeMBer_predict.M
    % predict mean and covariance of each Bernoulli component
    [SCB_MeMBer_predict.m{ii}, SCB_MeMBer_predict.P{ii}] = kalman_predict_multiple(model, SCB_MeMBer_update.m{ii}, SCB_MeMBer_update.P{ii});
end

% append birth Bernoulli components
for ii = 1:model.birth.J_gamma
    SCB_MeMBer_predict.r(SCB_MeMBer_predict.M + ii) = model.birth.r_gamma(ii);    

    SCB_MeMBer_predict.m{SCB_MeMBer_predict.M + ii} = model.birth.m_gamma(:,ii);
    SCB_MeMBer_predict.P{SCB_MeMBer_predict.M + ii} = model.birth.P_gamma;
    SCB_MeMBer_predict.w{SCB_MeMBer_predict.M + ii} = model.birth.mix_w_gamma(ii);
    SCB_MeMBer_predict.L(SCB_MeMBer_predict.M + ii) = length(model.birth.mix_w_gamma(ii));
end

% account for the increase in the Bernoulli components
SCB_MeMBer_predict.M = SCB_MeMBer_update.M + model.birth.J_gamma;

% clipping of existance probabilities
SCB_MeMBer_predict.r = limit_range(SCB_MeMBer_predict.r);
end

function SCB_MeMBer_sens_up = ms_update(SCB_MeMBer_predict, model, filter, Z)

SCB_MeMBer_sens_up = SCB_MeMBer_predict;

% Sequential update of the predcited multi-Bernoulli with the measurements
% form each sensor
for s_idx = 1:model.obs.nb_sensors
    % Perform CB-MeMBer update for sensor s_idx
    SCB_MeMBer_sens_up = ss_update(SCB_MeMBer_sens_up, model, filter, Z{s_idx}, s_idx);
    
    % Pruning of mixture components within each Bernoulli component 
    SCB_MeMBer_sens_up = mixture_merge_prune_cap(SCB_MeMBer_sens_up, filter.mix);     
end
    
end

function SCB_MeMBer_update = ss_update(SCB_MeMBer_predict, model, filter, Z, s_idx)

    %---construction of pseudo PHD for update
    L_pseudo = sum(SCB_MeMBer_predict.L);                                 %number of Gaussians in pseudo-PHD
    m_pseudo = zeros(model.kin.x_dim, L_pseudo);                          %means of Gaussians in pseudo-PHD
    P_pseudo = zeros(model.kin.x_dim, model.kin.x_dim, L_pseudo);         %covs of Gaussians in pseudo-PHD
    w_pseudo  = zeros(1, L_pseudo);                                       %weights of Gaussians in pseudo-PHD
    w_pseudo1 = zeros(1, L_pseudo);                                       %alt weight (1) of Gaussians in pseudo-PHD - used in CB-MeMBer update later
    w_pseudo2 = zeros(1, L_pseudo);                                       %alt weight (2) of Gaussians in pseudo-PHD - used in CB-MeMBer update later
    
    start_pt= 1;
    for t = 1:SCB_MeMBer_predict.M                  
        end_pt = start_pt+SCB_MeMBer_predict.L(t)-1;
        m_pseudo(:,start_pt:end_pt)  = SCB_MeMBer_predict.m{t};
        P_pseudo(:,:,start_pt:end_pt)= SCB_MeMBer_predict.P{t};
        w_pseudo(start_pt:end_pt) = SCB_MeMBer_predict.r(t)/(1-SCB_MeMBer_predict.r(t))*SCB_MeMBer_predict.w{t};
        w_pseudo1(start_pt:end_pt)= SCB_MeMBer_predict.r(t)/(1-SCB_MeMBer_predict.r(t)*model.obs.pD(s_idx))*SCB_MeMBer_predict.w{t};
        w_pseudo2(start_pt:end_pt)= SCB_MeMBer_predict.r(t)*(1-SCB_MeMBer_predict.r(t))/((1-SCB_MeMBer_predict.r(t)*model.obs.pD(s_idx))^2)*SCB_MeMBer_predict.w{t};
        start_pt = end_pt+1;
    end

    % number of measurements
    m = size(Z,2);
    % initialize updated multi-Bernoulli
    SCB_MeMBer_update.M = SCB_MeMBer_predict.M + m;                                    %total number of tracks/components
    SCB_MeMBer_update.L = zeros(1,SCB_MeMBer_update.M);
    SCB_MeMBer_update.r = zeros(1,SCB_MeMBer_update.M);
    SCB_MeMBer_update.w = cell(1,SCB_MeMBer_update.M);                                           
    SCB_MeMBer_update.m = cell(1,SCB_MeMBer_update.M);                                               
    SCB_MeMBer_update.P = cell(1,SCB_MeMBer_update.M);   
    
    %legacy tracks
    SCB_MeMBer_update.r(1:SCB_MeMBer_predict.M) = SCB_MeMBer_predict.r .* ((1-model.obs.pD(s_idx))./(1-SCB_MeMBer_predict.r*model.obs.pD(s_idx)));  %existence probability for tracks/components
    SCB_MeMBer_update.L(1:SCB_MeMBer_predict.M) = SCB_MeMBer_predict.L;                                                % number of mixture components for the legacy multi-Bernoulli
    SCB_MeMBer_update.w(1:SCB_MeMBer_predict.M) = SCB_MeMBer_predict.w;                                                % mixture weights
    SCB_MeMBer_update.m(1:SCB_MeMBer_predict.M) = SCB_MeMBer_predict.m;                                                % mean of track/component
    SCB_MeMBer_update.P(1:SCB_MeMBer_predict.M) = SCB_MeMBer_predict.P;                                                % covariance of track/component    
    
    %measurement updated tracks
    if m ~= 0
        offset = SCB_MeMBer_predict.M;

        gen_observation_s_fn = @(x) model.obs.gen_sen_obs(s_idx, x, []);

        [qz_update, m_update, P_update, ~, ~] = fast_ukf_update_multiple(Z, gen_observation_s_fn, m_pseudo, P_pseudo, model, filter.ukf);
        
        for z_idx = 1:m                                        
            r_temp = (model.obs.pD(s_idx)*qz_update(z_idx,:)*w_pseudo2(:))/(model.obs.lambda_c*model.obs.pdf_c+model.obs.pD(s_idx)*qz_update(z_idx,:)*w_pseudo1(:));
            w_temp = model.obs.pD(s_idx)*w_pseudo.*qz_update(z_idx,:) + eps; w_temp= w_temp/sum(w_temp);           
            
            SCB_MeMBer_update.r(offset+z_idx) = r_temp;
            SCB_MeMBer_update.L(offset+z_idx) = length(w_temp);
            
            SCB_MeMBer_update.m{offset+z_idx} = m_update(:,(1:length(w_temp)) + (z_idx-1)*length(w_temp));
            SCB_MeMBer_update.P{offset+z_idx} = P_update;
            SCB_MeMBer_update.w{offset+z_idx} = w_temp(:);                
        end      
        
    end

    SCB_MeMBer_update.r = limit_range(SCB_MeMBer_update.r);   

end % end function update


function member_out = prune(member_in, filter)
% prune Bernoulli components with weights below a given threshold
idx = find( member_in.r > filter.comp_threshold );

member_out.M = length(idx);
member_out.r = member_in.r(idx);
member_out.L = member_in.L(idx);

member_out.w = member_in.w(idx);
member_out.m = member_in.m(idx);
member_out.P = member_in.P(idx);

end % end function prune


function member_out = cap(member_in, N_cap)
% Cap total number of PHD components to a specified maximum

if member_in.M > N_cap
    [~, idxsort] = sort(member_in.r,'descend');
    idxkeep = idxsort(1:N_cap);

    member_out.M = length(idxkeep);
    member_out.r = member_in.r(idxkeep);
    member_out.L = member_in.L(idxkeep);

    member_out.w = member_in.w(idxkeep);
    member_out.m = member_in.m(idxkeep);
    member_out.P = member_in.P(idxkeep);
else
    member_out = member_in;
end

end % end function cap

function member_out = mixture_merge_prune_cap(member_in, filter_mix)
% Function that merges, prunes and caps the mixture components of each
% Bernoulli RFS of member_in.

member_out.M = member_in.M;
member_out.r = member_in.r;

%(within tracks) pruning,merging,capping
for t = 1:member_in.M
    [member_out.w{t}, member_out.m{t}, member_out.P{t}] = gaus_prune(member_in.w{t},member_in.m{t},member_in.P{t},filter_mix.elim_threshold);
    [member_out.w{t}, member_out.m{t}, member_out.P{t}] = gaus_merge(member_in.w{t},member_in.m{t},member_in.P{t},filter_mix.merge_threshold);
    [member_out.w{t}, member_out.m{t}, member_out.P{t}] = gaus_cap(member_in.w{t},member_in.m{t},member_in.P{t},  filter_mix.L_max);
    member_out.L(t)  = length(member_out.w{t});
end

end % end mixture_merge_prune_cap


function [X, N] = extract_estimates(member)
% State inference from the updated multi Bernoulli distribution
cdn_pmf = prod(1-member.r) * esf(member.r ./ (1-member.r));

[~, idx_max_cdn] = max(cdn_pmf);
map_cdn = idx_max_cdn-1;
N = min(member.M, map_cdn);   % estimate number of targets

[~, idx_max_tgt] = sort(member.r, 'descend');
X = [];
for t = 1:N % inferrence of component means
    [~,idx]= max(member.w{idx_max_tgt(t)});    % get mixture component with maximum mixture weight
    X(:,t) =  member.m{idx_max_tgt(t)}(:,idx);
end

end % end function extract estimates


function display_diaginfo( k, est, filter, H_predict, H_posterior, H_prune, H_cap)
% Display diagnostic information on the fly
if ~strcmp(filter.run_flag,'silence')
    disp([' time= ',num2str(k),...
          ' #est card=' num2str(est.state_n_hat(k),4),...
          ' #comp pred=' num2str(H_predict,4),...
          ' #comp post=' num2str(H_posterior,4),...
          ' #comp pruned=',num2str(H_prune),...
          ' #comp updt=',num2str(H_cap,4)   ]);
end
end % end function display