function output_cphd = filter_ms_cphd_kf(filter, model, meas, truth)
% output_cphd = filter_ms_cphd_kf(model, meas, truth)
% INPUT:
%      model - structure that contains kinematic and observation params
%      meas  - structure that contains measurements
%      truth - structure that contains ground truth tracks
% OUTPUT:
%      output_cphd - structure that hold output estimates
% -------------------------------------------------------------------------
% Function that implements the multi-sensor CPHD filter, proposed
% in [1]. The state models are linear and Gaussian, hence the 
% implementation is based on the Kalman filter. 
% -------------------------------------------------------------------------
% [1] S. Nannuru, S. Blouin, M. Coates, and M. Rabbat,  "Multisensor CPHD
% filter", IEEE Trans. Aerosp. and Electron. Syst., vol. 52, no. 4, pp. 
% 1834-1854, Aug. 2016, arXiv:1504.06342v2 [stat.ME]
%
% Contact: augustin.saucan@mail.mcgill.ca

% Jul 28 2017 - fix for selection of distinct partitions 


%=== Setup

%output variables
output_cphd.stat_X_hat = cell(meas.K,1);
output_cphd.state_n_hat = zeros(meas.K,1);
output_cphd.filter = filter;

cphd_mark_start = tic;      % start timing the filter

filter.fact_table = factorial(0:filter.max_card);
filter.nCi_table = zeros(filter.max_card+1);
for n = 0:filter.max_card
    index = 0:n;
    filter.nCi_table(n+1,index+1) = filter.fact_table(n+1) ./ (filter.fact_table(index+1) .* filter.fact_table(n-index+1));
end
filter.ps_powers = model.kin.pS .^ (0:filter.max_card);
filter.qs_powers = (1-model.kin.pS) .^ (0:filter.max_card);
filter.birth_cdn_pmf = poisspdf(0:filter.max_card, sum(model.birth.r_gamma));  % Poisson pmf for birth

%=== Filtering
%initial prior
MS_CPHD_update.J = 0;       % nb of components
MS_CPHD_update.w = [];      % weight of components
MS_CPHD_update.m = [];      % means of components
MS_CPHD_update.P = [];      % covariance of components
MS_CPHD_update.cdn_pmf = ones(1,filter.max_card+1)/(filter.max_card+1);

%recursive filtering
for k = 1:meas.K
    
    % Prediction and update
    MS_CPHD_predict = predict(MS_CPHD_update, model, filter);                                     H_predict = MS_CPHD_predict.J;
    MS_CPHD_update  = update(MS_CPHD_predict, model, filter, meas.Z{k});                          H_posterior = MS_CPHD_update.J;
    
    % Pruning & Merging
    MS_CPHD_update = prune_merge(MS_CPHD_update, filter);                                          H_prune = MS_CPHD_update.J;
       
    % State estimation
    [output_cphd.state_X_hat{k}, output_cphd.state_n_hat(k)] = extract_estimates(MS_CPHD_update);
    
    % Cap the PHD components at either the maximum number of allowed
    % components or to a number proportional with the number of estimated
    % targets
    N_cap = min([filter.comp_max, MS_CPHD_update.J, max(output_cphd.state_n_hat(k) * filter.nb_comp_tg, filter.nb_comp_tg) ]);
    MS_CPHD_update = cap(MS_CPHD_update, N_cap);                                                   H_cap = MS_CPHD_update.J;
     
    % Return the updated PHD component weights for diagnostics
    output_cphd.rkk{k} = MS_CPHD_update.w / sum(MS_CPHD_update.w);      
    
    % Display diagnostics
    display_diaginfo( k, output_cphd, filter, H_predict, H_posterior, H_prune, H_cap);
    
end

% return filter computation time 
output_cphd.all_time    = toc(cphd_mark_start);

% calculate and display average OSPA error
error_t = zeros(1,truth.K);
for k = 1:truth.K
    error_t(k) = ospa_dist( truth.X{k}, output_cphd.state_X_hat{k}, model.ospa_cutoff, model.ospa_order);
end

% return OSPA errors
output_cphd.ospa = error_t;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
%% %%%%%%%%%%%%%%%%%%%%%%%%%%% LOCAL FUNCTIONS %%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function MS_CPHD_predict = predict(MS_CPHD_update, model, filter)
% Predict PHD function
% Predict surviving tracks
MS_CPHD_predict = MS_CPHD_update;
MS_CPHD_predict.w = MS_CPHD_update.w * model.kin.pS;
if MS_CPHD_update.J
    [MS_CPHD_predict.m, MS_CPHD_predict.P] = kalman_predict_multiple(model, MS_CPHD_update.m, MS_CPHD_update.P);
end

%% Sort the surviving components in oder of decreasing weights. In the case of overlapping subsets for two predicted PHD components, the partition formation procedure assigns a measurement subset to the first predicted component and a miss detection for the second. Hence, it is better to have the highest component first.  
if MS_CPHD_predict.J
   [~, sorted] = sort(MS_CPHD_predict.w,'descend');

   MS_CPHD_predict.w= MS_CPHD_predict.w(sorted);
    
    for i = 1:MS_CPHD_predict.J
        sorted_m(:,i)   = MS_CPHD_predict.m(:,sorted(i));
        sorted_P(:,:,i) = MS_CPHD_predict.P(:,:,sorted(i));
    end
    MS_CPHD_predict.m = sorted_m;
    MS_CPHD_predict.P = sorted_P;  
end

% append birth PHD components
for i = 1:model.birth.J_gamma
    MS_CPHD_predict.w(MS_CPHD_predict.J + i)       = model.birth.r_gamma(i);
    MS_CPHD_predict.m(:,MS_CPHD_predict.J + i)     = model.birth.m_gamma(:,i);
    MS_CPHD_predict.P(:,:,MS_CPHD_predict.J + i)   = model.birth.P_gamma;
end

% account for the increase in the components
MS_CPHD_predict.J = MS_CPHD_update.J + model.birth.J_gamma;

% Predict cardinality point-mass function
pmf_kk1 = [];  % predicted cardinality

for n = 0:filter.max_card
    temp1 = 0;
    for j = 0:n
        index = j:filter.max_card;
        temp2 = filter.nCi_table(index+1,j+1)';
        temp2 = temp2 .* ( filter.ps_powers(j+1) * filter.qs_powers(index-j+1) );
        temp2 = temp2 * MS_CPHD_update.cdn_pmf(index+1)';
        
        temp1 = temp1 + filter.birth_cdn_pmf(n-j+1)*temp2;
    end
    pmf_kk1(n+1) = temp1;
end
MS_CPHD_predict.cdn_pmf = pmf_kk1/sum(pmf_kk1);
end

function MS_CPHD_update = update(MS_CPHD_predict, model, filter, Z)
% Update the predicted PHD function and cardinality distribution
% Check to see if the current measurement set is not empty
empty_data = 1;
for s = 1:model.obs.nb_sensors
    if ~isempty(Z{s})
        empty_data = 0;
        break;
    end
end

% Narmalize PHD function
MS_CPHD_predict.w = MS_CPHD_predict.w / sum(MS_CPHD_predict.w); 

gamma = prod(1-model.obs.pD);  % probability that all sensors miss the targets
    
if empty_data
    % if no sensor measurements are available, we keep the predicted PHD
    % components but with decreased wights.
    
    pgf_0 = gamma.^(0:filter.max_card) * MS_CPHD_predict.cdn_pmf'; % PGF of the predicted cardinality distribution
    pgf_1 = ((1:filter.max_card).*gamma.^(0:filter.max_card-1)) * MS_CPHD_predict.cdn_pmf(2:end)'; % first order derivative of the cardinality PGF
    
    ratio = (pgf_1/pgf_0)*gamma;  % (pgf_1/pgf_0) is actually alpha_0 for the void measurement set
    
    MS_CPHD_update.w = MS_CPHD_predict.w * ratio;    
    MS_CPHD_update.m = MS_CPHD_predict.m;
    MS_CPHD_update.P = MS_CPHD_predict.P;
    MS_CPHD_update.J = MS_CPHD_predict.J;
    MS_CPHD_update.cdn_pmf = MS_CPHD_predict.cdn_pmf;
       
else      
    %% Find best observation subsets (paths) for each PHD component
    [trellis_paths, trellis_weights, trellis_path_index, n_paths, m_paths, P_paths, rhoW, dW] = find_best_subsets(model, filter, MS_CPHD_predict, Z);
   
    %% Find likely partitions
    [~, partitions_v] = find_best_partitions(MS_CPHD_predict.J, trellis_paths, trellis_path_index, trellis_weights, n_paths, filter.P_max);
    
    % calculate path weights by summing over corresponding partitions
    % predicted pgf derivatives evaluated at gamma and calculation of psi_P
    psi(1) = gamma.^(0:filter.max_card) * MS_CPHD_predict.cdn_pmf';
    for i = 1:n_paths+1
        index = i:filter.max_card;
        nPi = filter.fact_table(index+1) ./ filter.fact_table(index - i + 1);
        gamma_powers = gamma.^(index - i);
        pgf_coeff = nPi.*gamma_powers;
        psi(i+1) = pgf_coeff * MS_CPHD_predict.cdn_pmf(index+1)';   % predicted pgf derivative
    end   
    
   % efficient calculation of path weights
    log_dW = log(dW' + realmin);    % fix to avoid -Inf terms in LOG
    partition_size = sum(partitions_v,2);
    temp = bsxfun(@times,partitions_v,log_dW);
    
    partition_dw = exp(sum(temp,2));
    partition_w = partition_dw .* psi(partition_size+1)';
    partition_w_star = partition_dw .* psi(partition_size+2)';

    temp = bsxfun(@times, partitions_v, partition_w);
    path_w = sum(temp,1)';
    path_w = path_w/sum(partition_w);

    for i = 1:n_paths
        for j = 1:MS_CPHD_predict.J
            rhoW((i-1)*MS_CPHD_predict.J+j) = rhoW((i-1)*MS_CPHD_predict.J+j)*path_w(i);
        end
    end

    % case of no target detection by any sensor (PHD update)
    ratio = (sum(partition_w_star)/sum(partition_w))*gamma;
    MS_CPHD_update.w = [rhoW MS_CPHD_predict.w*ratio];
    MS_CPHD_update.m = [m_paths MS_CPHD_predict.m];
    MS_CPHD_update.P = cat(3, P_paths, MS_CPHD_predict.P);
    MS_CPHD_update.J = MS_CPHD_predict.J * (1 + n_paths);
 
    % Update cardinality distribution
    beta = zeros(1,filter.max_card+1);
    pmf_kk = zeros(1,filter.max_card+1);
    partition_dw_size = zeros(1,n_paths+1);
    gamma_powers = gamma.^(0:filter.max_card);
    for j = 0:n_paths
        index = partition_size == j;
        partition_dw_size(j+1) = sum(partition_dw(index));
    end

    for n = 0:filter.max_card
        for j = 0:min(n,n_paths)
            beta(n+1) = beta(n+1) + partition_dw_size(j+1)*gamma_powers(n-j+1)*filter.fact_table(n+1)/filter.fact_table(n-j+1);
        end
        pmf_kk(n+1) = beta(n+1) * MS_CPHD_predict.cdn_pmf(n+1);
    end
    % normalization of the pmf
    MS_CPHD_update.cdn_pmf = pmf_kk/sum(pmf_kk);    
    
end  % if ~empty data
end % end function update

   

function [best_paths, best_paths_weights, best_paths_index, n_paths, m_paths, P_paths, w_paths, dW] = find_best_subsets(model, filter, MS_CPHD_predict, Z)
% Function that searches for the best W_max subsets for each PHD component

nb_sensors = model.obs.nb_sensors;  % number of sensors

J_kk1 = MS_CPHD_predict.J; % number of predicted PHD components

% initialize output variables
best_paths = cell(1,J_kk1);
best_paths_weights = cell(1,J_kk1);
dW = cell(1,J_kk1);

% for each component
for j = 1:J_kk1

    curr_J  = 1; % the number of current subsets W, starts with 1 empty W
    curr_paths = []; % the paths of all the current subsets W
    % we initilize the path parameters with the predicted parameters 
    curr_w  = MS_CPHD_predict.w(j) ; % we assume the mass of the current component as initial subset weight 
    curr_m  = MS_CPHD_predict.m(:,j);
    curr_P  = MS_CPHD_predict.P(:,:,j);
        
    % process each sensor sequentially
    for s = 1:model.obs.nb_sensors
        curr_Z = Z{s};
        nb_meas = size(curr_Z, 2);    % current number of measurements
            
         % Kalman updates for the for no detected observation case
        local_w = curr_w*(1 - model.obs.pD(s));
        local_m = curr_m;
        local_P = curr_P;
        
        if nb_meas ~= 0
            % Kalman filter updates
            [qz_update, m_update, P_update] = kalman_update_multiple(curr_Z, model, curr_m, curr_P);
            
            new_w = (model.obs.pD(s)/model.obs.pdf_c*repmat(curr_w, nb_meas,1) .* qz_update')';
            local_w = [local_w new_w(:)'];
            local_m = [local_m, m_update];
            local_P = cat(3, local_P, repmat(P_update, 1,1, nb_meas));
        end       
      
        local_paths = repmat(curr_paths, nb_meas+1, 1);
        new_obs     = kron(1:nb_meas, ones(curr_J, 1));
        new_obs     = [-1*ones(curr_J,1); new_obs(:)];
        local_paths = [local_paths new_obs];                      
        
        % select best observation sequences till now 
        % always include the empty observation path first
        [~,index] = sort(local_w(2:end),'descend');
        % select only the best W_max measurements as candidates for the j component
        J = min(length(index), filter.W_max);  % update number of components propagated
        index = index(1:J) + 1; % account for the void measurement association
        curr_w = [local_w(1) local_w(index)]; % first association is with the void measurement
        curr_m = [local_m(:,1) local_m(:,index)];
        curr_P = cat(3, local_P(:,:,1), local_P(:,:,index));
        curr_paths = [local_paths(1,:); local_paths(index,:)];
        curr_J = J + 1;
    end % for each sensor
    
    % final paths information for j-th Gaussian component
    best_paths{j} = curr_paths;
    
end % end for each PHD component

%% generate the collection of paths - all_paths
all_paths = [];
for j = 1:J_kk1
    [all_paths] = [all_paths; best_paths{j}];
end

% Getting rid of the line [-1 -1 ... -1] and any non-unique lines
all_paths = unique(all_paths, 'rows'); % due to unique, all_paths will be sorted and component information lost
index = sum(all_paths ~= -1,2) == 0; % eliminate rows of the type [-1 -1 ... -1]
all_paths(index,:) = [];
n_paths = size(all_paths,1); % nb of lines in all_paths

% best_paths_index only indicates which lines in all_paths correspond to
% the Gaussian component r.
% For l=1, we search for the [-1 -1 ... -1] line in all_paths.
% However, there are no such lines in all_paths and the
% best_paths_index{r}(1) = 0 for all r.
best_paths_index = cell(1,J_kk1);
for r = 1:J_kk1
    for l = 1:size(best_paths{r},1) 
        for i = 1:n_paths
            if all_paths(i,:) == best_paths{r}(l,:);
                best_paths_index{r}(l) = i;
                break;
            end
        end
    end
end

%% calculate path weights (rhoW and dW)

% we basically have n_paths subsets with paths given in all_paths

% initialize output GM components
w_paths   = repmat(MS_CPHD_predict.w, 1, n_paths);
m_paths   = repmat(MS_CPHD_predict.m, 1, n_paths);
P_paths   = repmat(MS_CPHD_predict.P, [1,1,n_paths]);


% initialize path variables
pathW = zeros(n_paths, 1);
dW = zeros(n_paths, 1);

for i = 1:n_paths
    temp_w1 = MS_CPHD_predict.w;  % temp variable for calculating path weights
    temp_w2 = MS_CPHD_predict.w;  % temp variable for calculating dW (with clutter_rate)
    % both temp_w1 and temp_w2 are vectors of length J_kk1

   % vectorize over all components J_kk1
   index = (i-1)*J_kk1 + (1:J_kk1);
    
    for s = 1:nb_sensors
          
        if all_paths(i,s) == -1 % void association for component i at sensor s
            w_paths(index) = w_paths(index)*(1-model.obs.pD(s));
            temp_w1     = temp_w1 * (1-model.obs.pD(s));
            temp_w2     = temp_w2 * (1-model.obs.pD(s));
        else
            % intermediate parameters
            z = Z{s}(:,all_paths(i,s));
            % Kalman filter updates
            [qz_update, m_update, P_update] = kalman_update_multiple(z, model, m_paths(:,index),  P_paths(:,:,index));            

            % update parameters (for rhoW and dW)
            w_paths(index) = model.obs.pD(s)*w_paths(index).*qz_update';
            temp_w1 = model.obs.pD(s)*temp_w1.*qz_update' / model.obs.pdf_c;
            temp_w2 = model.obs.pD(s)*temp_w2.*qz_update' / (model.obs.pdf_c * model.obs.lambda_c);
            % also update mW and sigmaW after each sensor measurement
            m_paths(:,index)   = m_update;
            P_paths(:,:,index) = P_update;
        end % if all_paths(i,s) 
    end % for each s

    % normalization
    normalization_factor = sum(w_paths(index));
    w_paths(index) = w_paths(index)/normalization_factor;    % rho_W x r_{k+1|k}
    pathW(i) = sum(temp_w1);    % dW (without clutter rate)
    dW(i) = sum(temp_w2);    % dW (with clutter rate)
end % for i = 1:n_paths (for all subsets W)

for r = 1:J_kk1
    % first path (the one for [-1 -1 ... -1]) has weight 1 by convention
    best_paths_weights{r} = ones(1,size(best_paths{r},1));
    for l = 2:size(best_paths{r},1) 
        best_paths_weights{r}(l) = pathW(best_paths_index{r}(l));
    end
end

end % end function find_best_subsets

function [n_partitions, partitions_v] = find_best_partitions(J_kk1, trellis_paths, trellis_path_index, trellis_weights, n_paths, max_partitions)                 
% Function that constructs the best partitions from the set of subsets
% returned by the find_best_subsets function.

n_partitions = 1;
partitions_list{1} = []; % we start with a single empty partition. The list of partitions will inrease after processing the subsets of each Gaussian.
unique_partitions = zeros(1,J_kk1);
w_partitions = 1;

for r = 1:J_kk1 % for each predicted Gaussian component
    n_temp = 0; w_temp = []; temp_list = {}; temp_up = [];
    
    for i = 1:n_partitions % each preavious partition will branck into severall new partitions by appending subsets of the r-th Gaussian component
        % trellis_paths{r} contains the W subsets of the r-th Gaussian component 
        for l = 1:size(trellis_paths{r},1) 
            % check to see if an observation contained in the current
            % trellis path trellis_paths{r}(l,:) is not already contained
            % in the partition partitions_list{i}.
            if ~overlap(partitions_list{i},trellis_paths{r}(l,:)) 
                n_temp = n_temp + 1; % number of partitions
                temp_up(n_temp,:) = unique_partitions(i,:);
                
                if trellis_path_index{r}(l) ~= 0 % if not the empty path
                    
                    % formation of a new partition by appending trellis_paths{r}(l,:)
                    % trellis_paths{r}(l,:) is a row of length s which contains the indices of observations (one for each of the s sensors) contained in the subset W 
                    temp_list{n_temp} = [partitions_list{i}; trellis_paths{r}(l,:)]; 
                    temp_up(n_temp,r) = trellis_path_index{r}(l); % remember which subset was added to this partition                   
                    w_temp(n_temp) = w_partitions(i)*trellis_weights{r}(l); % this represents the partition weight d_P = prod(d_W)
                else                    
                    % trellis_path_index{r}(l) = 0 signals that
                    % trellis_paths{r}(l,:) = [-1 -1 ... -1] (ie empty path)
                    % trellis_weights{r}(1) = 1 by convention
                    % partitions_list{i} is only necessary for checking the overlap. Overlap is not occouring for the void association -1, hence trellis_paths{r}(l,:) is not appended to temp list 
                    temp_list{n_temp} = partitions_list{i};                   
                    w_temp(n_temp) = w_partitions(i);             
                end
            end
        end
    end
    
    % keep only unique partitions
    temp_up = sort(temp_up,2,'descend');
    [temp_up,index,~] = unique(temp_up,'rows');
    temp_list = temp_list(index);
    w_temp = w_temp(index);
    n_temp = size(index,1);
    
    % keep partitions with highest weights
    n_partitions = min(max_partitions,n_temp);
    [~, index] = sort(w_temp,'descend');
    partitions_list = temp_list(index(1:n_partitions)); % update the partitions list
    unique_partitions = temp_up(index(1:n_partitions),:);
    w_partitions = w_temp(index(1:n_partitions));
end

% A binary vector representation of the partitions
partitions_v = zeros(n_partitions,n_paths);
for i = 1:n_partitions
    partitions_v(i,unique_partitions(i,unique_partitions(i,:)~=0)) = 1;
end

end  % end function find_best_partitions


function does_overlap = overlap(partition, path)
% function to verify if the observation subset 'path' overlaps with the 'partition'
does_overlap = 0;

p_size = size(partition,1);
nb_sensors = size(path,2);

p = 0;
while ~does_overlap && p < p_size
    p = p + 1;
    temp = partition(p,:);
    for s = 1:nb_sensors
        if (temp(s) ~= -1) && (path(s) ~= -1) && (temp(s) == path(s))
            does_overlap = 1;
            break;
        end
    end
end

end % end function overlap
 
function cphd_out = prune_merge(cphd_in, filter)
% prune UKF components with weights below a given threshold and merge
% close components. 


w_k = cphd_in.w;
m_k = cphd_in.m;
P_k = cphd_in.P;

ind = find(w_k>filter.comp_threshold);
merged = 0;
not_merged = ind;
ndim = size(m_k,1);
w_merge = []; m_merge = []; P_merge = [];
while ~isempty(not_merged)
    [~, ind_max] = max(w_k(not_merged));
    ind_max = not_merged(ind_max);
    ind_rest = not_merged(~(not_merged==ind_max));
    ind_merged = ind_max;
    
    P_to_invert = P_k(:,:,ind_max);
    symmetricP = mean(cat(3,P_to_invert,P_to_invert'),3);
    cholP= chol(symmetricP);
    cholInvP =  cholP \ eye(size(cholP));
    inverseP = cholInvP*cholInvP';
    
    m_minus = bsxfun(@minus,m_k(:,ind_max),m_k(:,ind_rest));
    mTPm = sum(m_minus'.*(inverseP*m_minus)',2); 
    % dot product above only computes the diagonal entries
    ind_merged = horzcat(ind_merged,ind_rest(mTPm<=filter.merge_threshold)); 
    
    merged = merged+1;
    w_merge(merged) = sum(w_k(ind_merged));
    m_merge(:,merged) = m_k(:,ind_merged)*w_k(ind_merged)'./w_merge(merged);
    m_diff = bsxfun(@minus,m_merge(:,merged), m_k(:,ind_merged));
    P_merge(:,:,merged) = zeros(ndim,ndim);
    for j = 1:length(ind_merged)
        P_merge(:,:,merged) = P_merge(:,:,merged) + ...
            ((P_k(:,:,ind_merged(j))+m_diff(:,j)*m_diff(:,j)') ...
            * w_k(ind_merged(:,j)));
    end
    P_merge(:,:,merged) = P_merge(:,:,merged)./w_merge(merged);
    not_merged = not_merged(~ismembc(not_merged,sort(ind_merged)));
end

% Prepare output

cphd_out.w = w_merge;
cphd_out.m = m_merge;
cphd_out.P = P_merge;
cphd_out.J = merged;
cphd_out.cdn_pmf = cphd_in.cdn_pmf;  %cardinality doesn't change


end % end function prune


function cphd_out = cap(cphd_in, N_cap)
% Cap total number of PHD components to a specified maximum

if cphd_in.J > N_cap
    [~, idxsort] = sort(cphd_in.w,'descend');
    idxkeep = idxsort(1:N_cap);

    cphd_out.J = length(idxkeep);
    cphd_out.w = cphd_in.w(idxkeep);
    cphd_out.cdn_pmf = cphd_in.cdn_pmf; %cardinality doesn't change
    
    for i = 1:length(idxkeep)
        cphd_out.m(:,i) = cphd_in.m(:,idxkeep(i));
        cphd_out.P(:,:,i) = cphd_in.P(:,:,idxkeep(i));
    end   
else
    cphd_out = cphd_in;
end

end % end function cap


function [X_hat, N_hat] = extract_estimates(cphd)
% State inference from the updated multi PHD distribution

% MAP cardinality estimate
[~, N_hat] = max(cphd.cdn_pmf);
N_hat = N_hat - 1;

[~, ind] = sort(cphd.w,'descend');
N_hat    = min(N_hat,length(ind));   % to avoid array access overflow
ind      = ind(1:N_hat);

X_hat = [];
for i = 1:N_hat
    % pick highest weight components as targets
    X_hat(:,i) = cphd.m(:,ind(i));
end

end % end function extract estimates


function display_diaginfo( k, est, filter, H_predict, H_posterior, H_prune, H_cap)
% Display diagnostic information on the fly
if ~strcmp(filter.run_flag,'silence')
    disp([' time= ',num2str(k),...
          ' #est card=' num2str(est.state_n_hat(k),4),...
          ' #comp pred=' num2str(H_predict,4),...
          ' #comp post=' num2str(H_posterior,4),...
          ' #comp pruned=',num2str(H_prune),...
          ' #comp updt=',num2str(H_cap,4)   ]);
end
end % end function display