function output_member = filter_ms_member_smc(filter, model, meas, truth)
% output_member = filter_ms_member_smc(model, meas, truth)
% INPUT:
%      model - structure that contains kinematic and observation params
%      meas  - structure that contains measurements
%      truth - structure that contains ground truth tracks
% OUTPUT:
%      output_member - structure that hold output estimates
% -------------------------------------------------------------------------
% Function that implements the multi-sensor multi-Bernoulli filter proposed
% in [1]. The state models are non-linear and the implementation is 
% based on particle filters (SMC). 
% -------------------------------------------------------------------------
% [1] A. Saucan, M. Coates, and M. Rabbat,  "A multi-sensor multi-Bernoulli 
% filter"  2016, arXiv:1609.05108 [stat.ME]
%
% Contact: augustin.saucan@mail.mcgill.ca

%=== Setup

%output variables
output_member.state_X_hat = cell(meas.K,1);   % inferred target states
output_member.state_n_hat = zeros(meas.K,1);  % estimated number of targets
output_member.filter      = filter;           % store filtering params

%=== Filtering
member_mark_start = tic;      % start timing the filter

% void initial prior
MS_MeMBer_update.M = 0;       % nb of Bernoulli components
MS_MeMBer_update.r = [];      % prob of existence of component
MS_MeMBer_update.x = {};      % particles of component
MS_MeMBer_update.w = {};      % wights of particles

%recursive filtering
for k = 1:meas.K
    
   
    % Prediction and update
    MS_MeMBer_predict = predict(MS_MeMBer_update, model, filter);                                     H_predict = MS_MeMBer_predict.M;
    MS_MeMBer_update  = update(MS_MeMBer_predict, model, filter, meas.Z{k});                          H_posterior = MS_MeMBer_update.M; 
    
    % Pruning 
    MS_MeMBer_update = prune(MS_MeMBer_update, filter);                                                H_prune = MS_MeMBer_update.M;    
    
    % State estimation
    [output_member.state_X_hat{k}, output_member.state_n_hat(k)] = extract_estimates(MS_MeMBer_update);

    % Cap the Bernoulli components at either the maximum number of allowed
    % components or to a number proportional with the number of estimated
    % targets
    N_cap = min([filter.comp_max, MS_MeMBer_update.M, max(output_member.state_n_hat(k) * filter.nb_comp_tg, filter.nb_comp_tg) ]);
    MS_MeMBer_update = cap(MS_MeMBer_update, N_cap);                                                   H_cap = MS_MeMBer_update.M;
    
    % Return the updated Bernoulli existence probabilities for diagnostics
    output_member.rkk{k} = MS_MeMBer_update.r ;  
    
    % Display diagnostics
    display_diaginfo( k, output_member, filter, H_predict, H_posterior, H_prune, H_cap);  
       
end

% return filter computation time 
output_member.all_time    = toc(member_mark_start);

% calculate and display average OSPA error
error_t = zeros(1,truth.K);
for k = 1:truth.K
    error_t(k) = ospa_dist( truth.X{k}, output_member.state_X_hat{k}, model.ospa_cutoff, model.ospa_order);
end

% return OSPA errors
output_member.ospa = error_t;

end % end function display

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
%% %%%%%%%%%%%%%%%%%%%%%%%%%%% LOCAL FUNCTIONS %%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function MS_MeMBer_predict = predict(MS_MeMBer_update, model, filter)

% Predict surviving tracks
MS_MeMBer_predict   = MS_MeMBer_update;
MS_MeMBer_predict.r = MS_MeMBer_update.r * model.kin.pS;
for i = 1:MS_MeMBer_update.M
    % predict particles of each Bernoulli component 
    MS_MeMBer_predict.x{i} = model.kin.prop_fn(MS_MeMBer_update.x{i}, 'noise');%   gen_newstate_fn(model, MS_MeMBer_update.x{i}, 'noise');  
end

%% Sort the surviving components in oder of decreasing weights. In the case of overlapping subsets for two predicted PHD components, the partition formation procedure assigns a measurement subset to the first predicted component and a miss detection for the second. Hence, it is better to have the highest component first.  
if MS_MeMBer_predict.M
   [~, sorted] = sort(MS_MeMBer_predict.r,'descend');

   MS_MeMBer_predict.r= MS_MeMBer_predict.r(sorted);
    
    for i = 1:MS_MeMBer_predict.M
        sorted_w{i}   = MS_MeMBer_predict.w{sorted(i)};
        sorted_x{i}   = MS_MeMBer_predict.x{sorted(i)};
    end
    MS_MeMBer_predict.w = sorted_w;
    MS_MeMBer_predict.x = sorted_x;  
end

% append birth Bernoulli components
for i = 1:model.birth.J_gamma
    MS_MeMBer_predict.r(MS_MeMBer_predict.M + i) = model.birth.r_gamma(i);
    MS_MeMBer_predict.x{MS_MeMBer_predict.M + i} = repmat(model.birth.m_gamma(:,i), 1, filter.nb_part) + model.birth.sqrt_P_gamma * randn(model.kin.x_dim, filter.nb_part);
    MS_MeMBer_predict.w{MS_MeMBer_predict.M + i} = 1/filter.nb_part * ones(1,filter.nb_part);
end

% account for the increase in the Bernoulli components
MS_MeMBer_predict.M = MS_MeMBer_update.M + model.birth.J_gamma;

end

function MS_MeMBer_update = update(MS_MeMBer_predict, model, filter, Z)

% Check to see if the current measurement set is not empty
empty_data = 1;
for s = 1:model.obs.nb_sensors
    if ~isempty(Z{s})
        empty_data = 0;
        break;
    end
end
    
gamma = prod(1-model.obs.pD);  % probability that all sensors miss the targets
    
if empty_data
    % if no sensor measurements are available, we keep the
    % predicted Bernoulli components but with decreased existence
    % probabilities. This step is identical to the single sensor
    % MemBer filter with a meta-seonsor having 1-p_D = gamma
    
    MS_MeMBer_update = MS_MeMBer_predict;
    MS_MeMBer_update.r = gamma * MS_MeMBer_predict.r ./ (1 - MS_MeMBer_predict.r +  MS_MeMBer_predict.r * gamma);
else
    
    
    %% Find best observation subsets (paths) for each Bernoulli component
   [trellis_paths, x_k, w_k, dW] = find_best_subsets( model, filter, MS_MeMBer_predict, filter.W_max, Z );
        
    %% Find likely partitions
    [~, unique_partitions, alpha_P] = find_best_partitions( MS_MeMBer_predict.M, trellis_paths, dW, filter.P_max, model.obs.lambda_c );
    
    r_kk = 0;
    x_kk = {};
    w_kk = {};
    
    index = 0;
    
    for j = 1:MS_MeMBer_predict.M                                          % for each predicted Bernoulli component                    
        % For each predicted Bernoulli component find the unique 
        subsets_j = unique_partitions(:,j);                                % select subsets assigned to the j-th Bernoulli component     
        [unique_subsets_j] = unique(subsets_j,'rows');                     % consider only the unique subsets
        
        nb_unique_subsets_j = length(unique_subsets_j);
        
        for  i = 1:nb_unique_subsets_j                                     % an updated Bernoulli component is created for each unique subset 

            index = index + 1;
            if unique_subsets_j(i) ~= 0                                               % if subset is not the empty subset \emptyset_{1:s}
                r_kk(index) = sum( alpha_P( subsets_j ==  unique_subsets_j(i) ) );    % add the weight of all partitions that assign the subset unique_subsets_j(i) to the j-th Bernoulli
                r_kk(index) = limit_range(r_kk(index));                               % clip the component prob. of existane to 0.999 (for numerical stability when calculating cardianlity pmf) 
                
                index_subsets = (unique_subsets_j(i)-1)*model.kin.x_dim + [1:model.kin.x_dim];
                
                x_kk{index} = x_k{j}(index_subsets, :);
                w_kk{index} = w_k{j}(unique_subsets_j(i), :);
            else                                                                       % if subset is the empty subset => legacy component, use the predicted Bernoulli and only update the existece probability
                r_kk(index) = sum( alpha_P( subsets_j ==  0 ) ) * gamma * MS_MeMBer_predict.r(j) ./ (1 - MS_MeMBer_predict.r(j) +  MS_MeMBer_predict.r(j) * gamma);
                r_kk(index) = limit_range(r_kk(index));
                
                x_kk{index} = MS_MeMBer_predict.x{j};                       % mean and covaraicne are unchanged
                w_kk{index} = MS_MeMBer_predict.w{j};
            end
        end
    end % for each of the predicted Bernoulli components
    
    % Return updated Bernoulli components
    MS_MeMBer_update.M = length(r_kk);
    MS_MeMBer_update.r = r_kk;
    MS_MeMBer_update.w = w_kk;
    MS_MeMBer_update.x = x_kk;
    
end  % if ~empty data

end % end function update

function [best_paths, x_paths, w_paths, dW] = find_best_subsets( model, filter, MS_MeMBer_predict, W_max, Z)
% Function that selects the best subsets for each predicted Bernoulli
% component.
nb_part = filter.nb_part;

M_k1 = MS_MeMBer_predict.M; % number of predicted Bernoulli components
gamma = prod(1-model.obs.pD);   % 1-p_D for the meta-sensor 

% initialize output variables
best_paths = cell(1,M_k1);
dW = cell(1,M_k1);

% for each Bernoulli component
for j = 1:M_k1
  
    curr_J  = 1; % the number of current subsets W, starts with 1 empty W
    curr_paths = []; % the paths of all the current subsets W
    % we initilize the path parameters with the predicted parameters 
    curr_r  = MS_MeMBer_predict.r(j); 
    curr_x  = MS_MeMBer_predict.x{j};
    curr_log_w  = log( curr_r * MS_MeMBer_predict.w{j} / sum(MS_MeMBer_predict.w{j}) );
    
     
    % process each sensor sequentially
    for s = 1:model.obs.nb_sensors
        curr_Z = Z{s};
        nb_meas = size(curr_Z,2);    % current number of measurements
        
        if nb_meas ~= 0
            % compute for each particle a corresponding predicted measurement
            curr_eta = model.obs.gen_sen_obs(s, curr_x, []);
            
            curr_eta_extend = repmat(curr_eta, 1, nb_meas);
            
            currZ_extend =  kron(curr_Z, ones(1, nb_part)); % tensor product to replicate the columns of curr_Z
            
            log_q = model.obs.lik_fn(currZ_extend, curr_eta_extend, model.obs.R)';  % compute log-likelihoods for each particle and each measurement of sensor s
            
            % size(log_q) = (mb_part * nb_meas) x 1
            %
            %               z^s_1         z^s_2                 z^s_m
            % log_q' = |______________|______________| ... |______________|
            %          1     nb_part  1       nb_part      1       nb_part
            
            % extend the the matrix of likelihood values
            log_q_extend = repmat(log_q.',curr_J, 1);  % size(log_q_extended) = curr_J x (nb_part * nb_meas)
            
            % Update the particle weights of the curr_J subsets with the new
            % observation likelihoods and p_D and clutter_density.
            curr_log_w_ex = log_q_extend + repmat(curr_log_w, 1, nb_meas) + ...
                kron(  log(model.obs.pD(s)/model.obs.pdf_c) * ones( curr_J, nb_meas),ones(1,nb_part));
            
            % Reshape curr_log_w_ex in order to match local_paths.
            % size(local_log_w) = (curr_J * nb_meas) x nb_part
            % Each block of curr_J rows of local_log_w correspond to the
            % particle weights of the curr_J subsets updated with the new
            % observations.
            out = reshape(permute( reshape(curr_log_w_ex, curr_J, nb_part,[])  , [1 3 2]), curr_J*nb_meas,[] );
        else
            out = [];
        end
        
        % Updates for the for no detected observation case
        local_log_w = log(1- model.obs.pD(s)) + curr_log_w;
        % Finally append non-detected subsets to the measurement updated
        % subsets.
        local_log_w  = [local_log_w;  out];
        
        % Compute the wights of the newly formed subsets
        local_r = sum(exp(local_log_w),2);
        local_r = local_r';
        
        % Local paths will contain a vector for the associations between
        % gaussian components and the measurements. First element is always
        % -1 for the void measurement. Additional elements will be indexed
        % by the index of each individual measurement in the measurement set of sensor
        % s. The general shape of local_path is given next:
        
        %                |                |   -1  |
        %                |  [curr_paths]  |   -1  |
        %                |________________|   -1  |
        %                |                | z^s_1 |
        %                |  [curr_paths]  | z^s_1 |
        %  local_paths = |________________| z^s_1 |
        %                        .
        %                        .
        %                        .
        %                |                | z^s_m |
        %                |  [curr_paths]  | z^s_m |
        %                |________________| z^s_m |
        
        
        local_paths = repmat(curr_paths, nb_meas+1, 1);
        new_obs = kron(1:nb_meas, ones(curr_J, 1));
        new_obs = [-1*ones(curr_J,1); new_obs(:)];
        local_paths = [local_paths new_obs];
        
        % select best observation sequences till now
        % always include the empty observation path first
        [~,index] = sort(local_r(2:end),'descend');
        % select only the best W_max measurements as candidates for the j component
        J = min(length(index),W_max);  % the number of non-void subsets
        index = index(1:J) + 1; % select best subsets and account for the void measurement subset
        
        % size(local_log_w) = (curr_J * nb_meas) x nb_part
        
        curr_log_w = [local_log_w(1,:); local_log_w(index,:)];
        
        curr_paths = [local_paths(1,:); local_paths(index,:)];
        curr_J = J + 1;  % save current number of subsets
    end % for each sensor
    
    % final paths information for j_th Bernoulli component
    % each column of curr_paths corresponds to a sensor while the lines are different combinations of the sensor measurements 
    best_paths{j} = curr_paths;
    
    
    x_paths{j} = [];  %  initialise the resampled particles
    
    curr_w = exp(curr_log_w);  % particle wights
    curr_r = sum(curr_w, 2)';  % weights of subsets (also serves to normalize the particle cloud of each subset)
    
    for i = 1:curr_J  % resample for each of the selected subsets
        w_paths{j}(i,:)    = curr_w(i,:) / curr_r(i); % normalize particles
        nb_eff_samp = 1/sum((w_paths{j}(i,:)).^2);    % effective sample size
        
        if nb_eff_samp <= (filter.nb_eff_part)             % resampling
            [ X_tilde, w_paths{j}(i,:), ~ ] = Resampling( curr_x, w_paths{j}(i,:) , nb_part, 'stratified' ); % resample particles
            x_paths{j}   = [x_paths{j}; X_tilde]; % store resampled particles
        else % no resampling is needed
            x_paths{j}   = [x_paths{j}; curr_x];
        end
    end
        
    % compute final weights of subsets
    curr_r(1) = (1 - MS_MeMBer_predict.r(j) + MS_MeMBer_predict.r(j) * gamma);
    dW{j}     = curr_r;
end % end for each Bernoulli component

end % end function

function [n_partitions, unique_partitions, alpha_P] = find_best_partitions(M_k1, trellis_paths, dW, max_partitions, clutter_rate)
% Function that constructs the best partitions from theselected subsets

n_partitions = 1;  % total number of partitions
partitions_list{1} = []; % we start with a single empty partition. The list of partitions will inrease after processing the subsets of each Bernoulli.
unique_partitions = zeros(1,M_k1);
log_w_partitions = 0;

for r = 1:M_k1 % for each predicted Bernoulli component
    n_temp = 0; % local variable for the number of partitions 
    w_temp = []; 
    temp_list = {}; 
    temp_up = [];
    
    for i = 1:n_partitions % each preavious partition will branck into severall new partitions by appending subsets of the r-th Bernoulli component
        % trellis_paths{r} contains the W subsets of the r-th Bernoulli component 
        for l = 1:size(trellis_paths{r},1) 
            % check to see if an observation contained in the current
            % trellis path trellis_paths{r}(l,:) is not already contained
            % in the partition partitions_list{i}.
            if ~overlap(partitions_list{i},trellis_paths{r}(l,:))
                n_temp = n_temp + 1; % number of partitions
                temp_up(n_temp,:) = unique_partitions(i,:);
                
                if sum(trellis_paths{r}(l,:) == -1) ~=  length(trellis_paths{r}(l,:))   % if not the empty path trellis_path_index{r}(l) ~= 0
                    
                    % formation of a new partition by appending trellis_paths{r}(l,:)
                    % trellis_paths{r}(l,:) is a row of length s which contains the indices of observations (one for each of the s sensors) contained in the subset W 
                    temp_list{n_temp} = [partitions_list{i}; trellis_paths{r}(l,:)]; 
                    temp_up(n_temp,r) = l; % remember which subset was added to this partition trellis_path_index{r}(l)                 
                else                    
                    % trellis_path_index{r}(l) = 0 signals that
                    % trellis_paths{r}(l,:) = [-1 -1 ... -1] (ie empty path)
                    temp_list{n_temp} = partitions_list{i};                                   
                end
                w_temp(n_temp) =  log_w_partitions(i) + log( dW{r}(l) ); % this represents the partition weight d_P = prod(d_W)                   
            end % if overlap
        end % for l each subset W associated with the Bernoulli component r
    end % for r each of the previous partitions 
    
    % keep only unique partitions
    [temp_up,index,~] = unique(temp_up,'rows');
    temp_list = temp_list(index);
    w_temp = w_temp(index);
    n_temp = size(index,1);
    
    % keep partitions with highest weights
    n_partitions = min(max_partitions,n_temp);
    [~, index] = sort(w_temp,'descend');
    partitions_list = temp_list(index(1:n_partitions)); % update the partitions list
    unique_partitions = temp_up(index(1:n_partitions),:);
    log_w_partitions = w_temp(index(1:n_partitions));
end  % for each Bernoulli component

% compute partition weights alpha_{P,J,B}
log_alpha_P = zeros(size(log_w_partitions));
for i =1:n_partitions    
    nb_obs = sum(sum( partitions_list{i} ~= -1 ));    
    log_alpha_P(i) = log_w_partitions(i) + log( clutter_rate^(-nb_obs) );      
end

 alpha_P = exp( log_alpha_P - log( sum( exp(log_alpha_P) ) ) ); % normalize
end

function does_overlap = overlap(partition,path)
% function to find if the observation subset 'path' overlaps with the 'partition'

does_overlap = 0;

p_size = size(partition,1);
no_sensors = size(path,2);

p = 0;
while ~does_overlap && p < p_size
    p = p + 1;
    temp = partition(p,:);
    for s = 1:no_sensors
        if (temp(s) ~= -1) && (path(s) ~= -1) && (temp(s) == path(s))
            does_overlap = 1;
            break;
        end
    end
end

end

function member_out = prune(member_in, filter)
% prune Bernoulli components with weights below a given threshold
idx = find( member_in.r > filter.comp_threshold );

member_out.M = length(idx);
member_out.r = member_in.r(idx);

for i = 1:length(idx)
    member_out.x{i} = member_in.x{idx(i)};
    member_out.w{i} = member_in.w{idx(i)};
end
end % end function prune


function member_out = cap(member_in, N_cap)
% Cap total number of Bernoulli components to a specified maximum

if member_in.M > N_cap
    [~, idxsort] = sort(member_in.r,'descend');
    idxkeep = idxsort(1:N_cap);

    member_out.M = length(idxkeep);
    member_out.r = member_in.r(idxkeep);
    for i = 1:length(idxkeep)
        member_out.x{i} = member_in.x{idxkeep(i)};
        member_out.w{i} = member_in.w{idxkeep(i)};
    end   
else
    member_out = member_in;
end

end % end function cap


function [X, N] = extract_estimates(member)
% State inference from the updated multi Bernoulli distribution
cdn_pmf = prod(1-member.r) * esf(member.r ./ (1-member.r));

[~, idx_max_cdn] = max(cdn_pmf);
map_cdn = idx_max_cdn-1;
N = min(member.M, map_cdn);   % estimate number of targets

[~, idx_max_tgt] = sort(member.r, 'descend');
X = [];
for t = 1:N % compute sample means for each N estimated targets
    X(:,t) =  member.x{idx_max_tgt(t)} * member.w{idx_max_tgt(t)}(:);
end

end % end function extract estimates


function display_diaginfo( k, est, filter, H_predict, H_posterior, H_prune, H_cap)
% Display diagnostic information on the fly
if ~strcmp(filter.run_flag,'silence')
    disp([' time= ',num2str(k),...
          ' #est card=' num2str(est.state_n_hat(k),4),...
          ' #comp pred=' num2str(H_predict,4),...
          ' #comp post=' num2str(H_posterior,4),...
          ' #comp pruned=',num2str(H_prune),...
          ' #comp updt=',num2str(H_cap,4)   ]);
end
end % end function display
