function output_tcphd = filter_ms_tcphd_smc(filter, model, meas, truth)
% output_cphd = filter_ms_tcphd_smc(model, meas, truth)
% INPUT:
%      model - structure that contains kinematic and observation params
%      meas  - structure that contains measurements
%      truth - structure that contains ground truth tracks
% OUTPUT:
%      output_tcphd - structure that hold output estimates
% -------------------------------------------------------------------------
% Function that implements the multi-sensor TCPHD filter [1] (a modified 
% version of the MS-CPHD filter of [2]). The state models are non-linear 
% and the implementation is based on particle filters (SMC). 
% -------------------------------------------------------------------------
% [1] A. Saucan, M. Coates, and M. Rabbat,  "A multi-sensor multi-Bernoulli 
% filter", 2016, arXiv:1609.05108 [stat.ME]
% [2] S. Nannuru, S. Blouin, M. Coates, and M. Rabbat, "Multisensor CPHD
% filter", IEEE Trans. Aerosp. and Electron. Syst., vol. 52, no. 4, pp. 
% 1834-1854, Aug. 2016, arXiv:1504.06342v2 [stat.ME]
%
% Contact: augustin.saucan@mail.mcgill.ca

%=== Setup

%output variables
output_tcphd.state_X_hat = cell(meas.K,1);
output_tcphd.state_n_hat = zeros(meas.K,1);
output_tcphd.filter = filter;

cphd_mark_start = tic;      % start timing the filter

% precomputations
filter.fact_table = factorial(0:filter.max_card);
filter.nCi_table = zeros(filter.max_card+1);
for n = 0:filter.max_card
    index = 0:n;
    filter.nCi_table(n+1,index+1) = filter.fact_table(n+1) ./ (filter.fact_table(index+1) .* filter.fact_table(n-index+1));
end
filter.ps_powers = model.kin.pS .^ (0:filter.max_card);
filter.qs_powers = (1-model.kin.pS) .^ (0:filter.max_card);
filter.birth_cdn_pmf = poisspdf(0:filter.max_card, sum(model.birth.r_gamma));  % Poisson pmf for birth

%=== Filtering

%initial prior
MS_TCPHD_update.J = 0;       % nb of particle clouds, i.e. components
MS_TCPHD_update.r = [];      % weight of component
MS_TCPHD_update.x = {};      % particles of component
MS_TCPHD_update.w = {};      % pdf-wights of a component
MS_TCPHD_update.cdn_pmf = ones(1,filter.max_card+1)/(filter.max_card+1);

%recursive filtering
for k = 1:meas.K
         
    % Prediction and update
    MS_TCPHD_predict = predict(MS_TCPHD_update, model, filter);                                     H_predict = MS_TCPHD_predict.J;      
    MS_TCPHD_update  = update(MS_TCPHD_predict, model, filter, meas.Z{k});                          H_posterior = MS_TCPHD_update.J;
    
    % Pruning & Merging
    MS_TCPHD_update = prune_merge(MS_TCPHD_update, filter, model);                                  H_prune = MS_TCPHD_update.J;
       
    % State estimation
    [output_tcphd.state_X_hat{k}, output_tcphd.state_n_hat(k)] = extract_estimates(MS_TCPHD_update);
    
    % Cap the PHD components at either the maximum number of allowed
    % components or to a number proportional with the number of estimated
    % targets
    N_cap = min([filter.comp_max, MS_TCPHD_update.J, max(output_tcphd.state_n_hat(k) * filter.nb_comp_tg, filter.nb_comp_tg) ]);
    MS_TCPHD_update = cap_tracks(MS_TCPHD_update, N_cap);       H_cap = MS_TCPHD_update.J;
      
    % Return the updated PHD component weights for diagnostics
    output_tcphd.rkk{k} = MS_TCPHD_update.r / sum(MS_TCPHD_update.r);    
    
    % Display diagnostics
    display_diaginfo( k, output_tcphd, filter, H_predict, H_posterior, H_prune, H_cap);
    
end

% return filter computation time 
output_tcphd.all_time    = toc(cphd_mark_start);

% calculate and display average OSPA error
error_t = zeros(1,truth.K);
for k = 1:truth.K
    error_t(k) = ospa_dist( truth.X{k}, output_tcphd.state_X_hat{k}, model.ospa_cutoff, model.ospa_order);
end

% return OSPA errors
output_tcphd.ospa = error_t;

end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
%% %%%%%%%%%%%%%%%%%%%%%%%%%%% LOCAL FUNCTIONS %%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function MS_TCPHD_predict = predict(MS_TCPHD_update, model, filter)
% Predict PHD function

    if MS_TCPHD_update.J
        [~, sorted] = sort(MS_TCPHD_update.r,'descend');
        % sorted = 1:MS_TCPHD_update.J;
        MS_TCPHD_update.r = MS_TCPHD_update.r(sorted);
        
        for i = 1:MS_TCPHD_update.J
            new_MS_TCPHD_update.w{i} = MS_TCPHD_update.w{sorted(i)};
            new_MS_TCPHD_update.x{i} = MS_TCPHD_update.x{sorted(i)};
        end
        MS_TCPHD_update.w = new_MS_TCPHD_update.w;
        MS_TCPHD_update.x = new_MS_TCPHD_update.x;
    end

% Predict surviving tracks
MS_TCPHD_predict   = MS_TCPHD_update;
MS_TCPHD_predict.r = MS_TCPHD_update.r * model.kin.pS;

for i = 1:MS_TCPHD_update.J
     MS_TCPHD_predict.x{i} = model.kin.prop_fn(MS_TCPHD_update.x{i}, 'noise');
end

% Sort the surviving components in oder of decreasing weights. 
if MS_TCPHD_predict.J
   [~, sorted] = sort(MS_TCPHD_predict.r,'descend');

   MS_TCPHD_predict.r = MS_TCPHD_predict.r(sorted);
    
    for i = 1:MS_TCPHD_predict.J
        sorted_w{i} = MS_TCPHD_predict.w{sorted(i)};
        sorted_x{i} = MS_TCPHD_predict.x{sorted(i)};
    end
    MS_TCPHD_predict.w = sorted_w;
    MS_TCPHD_predict.x = sorted_x;  
end

% append birth PHD components
for i = 1:model.birth.J_gamma
    MS_TCPHD_predict.r(MS_TCPHD_predict.J + i) = model.birth.r_gamma(i);
    MS_TCPHD_predict.x{MS_TCPHD_predict.J + i} = repmat(model.birth.m_gamma(:,i), 1, filter.nb_part) + model.birth.sqrt_P_gamma * randn(model.kin.x_dim, filter.nb_part);
    MS_TCPHD_predict.w{MS_TCPHD_predict.J + i} = 1/filter.nb_part * ones(1,filter.nb_part);
end

% account for the increase in the components
MS_TCPHD_predict.J = MS_TCPHD_predict.J + model.birth.J_gamma;

% Predict cardinality point-mass function
pmf_kk1 = [];  % predicted cardinality

for n = 0:filter.max_card
    temp1 = 0;
    for j = 0:n
        index = j:filter.max_card;
        temp2 = filter.nCi_table(index+1,j+1)';
        temp2 = temp2 .* ( filter.ps_powers(j+1) * filter.qs_powers(index-j+1) );
        temp2 = temp2 * MS_TCPHD_update.cdn_pmf(index+1)';
        
        temp1 = temp1 + filter.birth_cdn_pmf(n-j+1)*temp2;
    end
    pmf_kk1(n+1) = temp1;
end
MS_TCPHD_predict.cdn_pmf = pmf_kk1/sum(pmf_kk1);
end

function MS_TCPHD_update = update(MS_TCPHD_predict, model, filter, Z)
% Update the predicted PHD function and cardinality distribution

MS_TCPHD_predict.r = MS_TCPHD_predict.r / sum(MS_TCPHD_predict.r);

% Check to see if the current measurement set is not empty
empty_data = 1;
for s = 1:model.obs.nb_sensors
    if ~isempty(Z{s})
        empty_data = 0;
        break;
    end
end

gamma = prod(1-model.obs.pD);  % probability that all sensors miss the targets
    
if empty_data
    % if no sensor measurements are available, we keep the predicted PHD
    % components but with decreased wights.
    
    pgf_0 = gamma.^(0:filter.max_card) * MS_TCPHD_predict.cdn_pmf'; % PGF of the predicted cardinality distribution
    pgf_1 = ((1:filter.max_card).*gamma.^(0:filter.max_card-1)) * MS_TCPHD_predict.cdn_pmf(2:end)'; % first order derivative of the cardinality PGF
    
    ratio = (pgf_1/pgf_0)*gamma;  % (pgf_1/pgf_0) is actually alpha_0 for the void measurement set
    
    MS_TCPHD_update.r = MS_TCPHD_predict.r * ratio;    
    MS_TCPHD_update.x = MS_TCPHD_predict.x;
    MS_TCPHD_update.w = MS_TCPHD_predict.w;
    MS_TCPHD_update.J = MS_TCPHD_predict.J;
    MS_TCPHD_update.cdn_pmf = MS_TCPHD_predict.cdn_pmf;
else      
    %% Find best observation subsets (paths) for each PHD component
    [trellis_paths, trellis_weights, n_paths, x_paths, w_paths] = find_best_subsets(model, filter, MS_TCPHD_predict, filter.W_max, Z);
       
    %% Find likely partitions
    [n_partitions, unique_partitions, kappa_P] = find_best_partitions(MS_TCPHD_predict.J, trellis_paths, trellis_weights, filter.P_max, model.obs.lambda_c);

    
    % predicted pgf derivatives evaluated at gamma and calculation of psi_P
    psi(1) = gamma.^(0:filter.max_card) * MS_TCPHD_predict.cdn_pmf';
    for i = 1:n_paths+1
        index = i:filter.max_card;
        nPi = filter.fact_table(index+1) ./ filter.fact_table(index - i + 1);
        gamma_powers = gamma.^(index - i);
        pgf_coeff = nPi.*gamma_powers;
        psi(i+1) = pgf_coeff * MS_TCPHD_predict.cdn_pmf(index+1)';   % predicted pgf derivative
    end    
    
    unique_partitions( unique_partitions == 0 ) = 1;
    
     % calculate path weights by summing over corresponding partitions
    d_P = zeros(n_partitions,1);
    for i = 1:MS_TCPHD_predict.J
        
        log_trellis_weights_i =  log(trellis_weights{i} + realmin);  % fix to avoid -Inf terms in LOG        
        for j = 1:n_partitions
            subsets_ij =   unique_partitions(j,i);
            d_P(j) = d_P(j) + log_trellis_weights_i(subsets_ij) * (subsets_ij ~= 1);  % do not consider void subsets
        end
    end
    
    d_P =  exp(d_P);
    
    partition_size = sum(unique_partitions ~= 1, 2); % number of non-void subsets in each partition
    
    % compute the alpha_P and alpha0
    partition_dW     = kappa_P' .* d_P;                                    % partition_dW is kappa_p * \prod_{W \in P} d_W from eq (21 -22) of [1]
    partition_w      = partition_dW .* psi(partition_size+1)';             % psi(partition_size+1) is  M^(|P|-1) from  eq (21 -22) of [1]
    partition_w_star = partition_dW .* psi(partition_size+2)';             % psi(partition_size+2) is  M^(|P|) 
    alpha0 = (sum(partition_w_star)/sum(partition_w))*gamma;               % alpha0 is \alpha_0 of eq. (21) of [1]
    alpha_P = partition_w / sum(partition_w);                              % alpha_P is \alpha_P of eq. (22) of [1]
    
    
    % case of no target detection by any sensor (PHD update)
    MS_TCPHD_update.x = MS_TCPHD_predict.x;  % particles of legacy components 
    MS_TCPHD_update.w = MS_TCPHD_predict.w;  % particle weights of legacy components  
    MS_TCPHD_update.J = MS_TCPHD_predict.J;
    
    MS_TCPHD_update.r = MS_TCPHD_predict.r * alpha0; % weights of legacy components (i.e., first term in eq. (23) of [1])   
    
    % compute weights, particles and particle weights for the updated components
    for j = 1:MS_TCPHD_predict.J % for each of the predicted components
        
        subsets_j = unique_partitions(:,j);
        [unique_subsets_j] = unique(subsets_j,'rows');
        
        nb_unique_subsets_j = length(unique_subsets_j);
        
        for  i = 1:nb_unique_subsets_j
            if unique_subsets_j(i) ~= 1  % not a void component
                MS_TCPHD_update.J = MS_TCPHD_update.J + 1;
                
                MS_TCPHD_update.r = [MS_TCPHD_update.r sum( alpha_P( subsets_j ==  unique_subsets_j(i) ) )];
                
                index_subsets = (unique_subsets_j(i)-1)*model.kin.x_dim + [1:model.kin.x_dim];
                MS_TCPHD_update.x{MS_TCPHD_update.J} = x_paths{j}(index_subsets,:);
                MS_TCPHD_update.w{MS_TCPHD_update.J} = w_paths{j}(unique_subsets_j(i),:);
            end
        end
    end % for each of the predicted PHD components
    
    % Update cardinality distribution
    beta = zeros(1,filter.max_card+1);
    pmf_kk = zeros(1,filter.max_card+1);
    partition_dw_size = zeros(1,n_paths+1);
    gamma_powers = gamma.^(0:filter.max_card);
    for j = 0:n_paths
        index = partition_size == j;
        partition_dw_size(j+1) = sum(partition_dW(index));
    end

    for n = 0:filter.max_card
        for j = 0:min(n,n_paths)
            beta(n+1) = beta(n+1) + partition_dw_size(j+1)*gamma_powers(n-j+1)*filter.fact_table(n+1)/filter.fact_table(n-j+1);
        end
        pmf_kk(n+1) = beta(n+1) * MS_TCPHD_predict.cdn_pmf(n+1);
    end
    % normalization of the pmf
    MS_TCPHD_update.cdn_pmf = pmf_kk/sum(pmf_kk);
    
end  % if ~empty data
end % end function update


function [best_paths, best_paths_weights, n_paths, x_paths, w_paths] = ...
         find_best_subsets(model, filter, MS_TCPHD_predict, W_max, Z)
  
% Function that searches for the best W_max subsets for each PHD component
nb_sensors = model.obs.nb_sensors;  % number of sensors
nb_part = filter.nb_part;

J_kk1 = MS_TCPHD_predict.J; % number of predicted PHD components

% initialize output variables
best_paths = cell(1,J_kk1);
best_paths_weights = cell(1,J_kk1);

% for each component
for j = 1:J_kk1

    curr_J  = 1; % the number of current subsets W, starts with 1 empty W
    curr_paths = []; % the paths of all the current subsets W
    % we initilize the path parameters with the predicted parameters 
    curr_r  = MS_TCPHD_predict.r(j) ; % we assume the mass of the current component as initial subset weight 
    curr_x  = MS_TCPHD_predict.x{j};
    curr_log_w  = log( curr_r * MS_TCPHD_predict.w{j}/sum(MS_TCPHD_predict.w{j}) ); % make sure w_k1 is a true pdf
        
    % process each sensor sequentially
    for s = 1:nb_sensors
        curr_Z = Z{s};
        nb_meas = size(curr_Z, 2);    % current number of measurements
        
        if nb_meas ~= 0
            % compute for each particle a corresponding predicted measurement
            curr_eta = model.obs.gen_sen_obs(s, curr_x, []);
            
            curr_eta_extend = repmat(curr_eta, 1, nb_meas);
            
            currZ_extend =  kron(curr_Z, ones(1, nb_part)); % tensor product to replicate the columns of curr_Z
            
            log_q = model.obs.lik_fn(currZ_extend, curr_eta_extend, model.obs.R)';
            
            % size(log_q) = (nb_part(j) * nb_meas) x 1
            %
            %                z^s_1         z^s_2                 z^s_m
            % log_q' = |______________|______________| ... |______________|
            %          1     nb_part  1       nb_part      1       nb_part
            
            % extend the the matrix of likelihood values
            log_q_extend = repmat(log_q.', curr_J, 1);  % size(log_q_extended) = curr_J x (nb_part(j) * nb_meas)
            
            % Update the particle weights of the curr_J subsets with the new
            % observation likelihoods and p_D and clutter_density.
            curr_log_w_ex = log_q_extend + repmat(curr_log_w, 1, nb_meas) + ...
                kron( log(model.obs.pD(s)/model.obs.pdf_c) * ones(curr_J, nb_meas), ones(1, nb_part) );
            
            % Reshape curr_log_w_ex in order to match local_paths.
            % size(local_log_w) = (curr_J * nb_meas) x nb_part(j)
            % Each block of curr_J rows of local_log_w correspond to the
            % particle weights of the curr_J subsets updated with the new
            % observations.
            out = reshape(permute( reshape(curr_log_w_ex, curr_J, nb_part,[]), [1 3 2]), curr_J*nb_meas,[] );
        else
            out = [];
        end
        
        % Updates for the for no detected observation case
        local_log_w = log(1 - model.obs.pD(s)) + curr_log_w;
        % Finally append non-detected subsets to the measurement updated
        % subsets.
        local_log_w  = [local_log_w;  out];
        
        % Compute the wights of the newly formed subsets
        local_r = sum(exp(local_log_w),2);
        local_r = local_r';
        
        % Local paths will contain a vector for the associations between
        % PHD components and the measurements. First element is always
        % -1 for the void measurement. Additional elements will be indexed
        % by the index of each individual measurement in the measurement set of sensor
        % s. The general shape of local_path is given next:
        
        %                |                |   -1  |
        %                |  [curr_paths]  |   -1  |
        %                |________________|   -1  |
        %                |                | z^s_1 |
        %                |  [curr_paths]  | z^s_1 |
        %  local_paths = |________________| z^s_1 |
        %                        .
        %                        .
        %                        .
        %                |                | z^s_m |
        %                |  [curr_paths]  | z^s_m |
        %                |________________| z^s_m |
        
        
        local_paths = repmat(curr_paths, nb_meas+1, 1);
        new_obs = kron(1:nb_meas, ones(curr_J, 1));
        new_obs = [-1*ones(curr_J,1); new_obs(:)];
        local_paths = [local_paths new_obs];
        
        % select best observation sequences till now
        % always include the empty observation path first
        [~,index] = sort(local_r(2:end),'descend');
        % select only the best W_max measurements as candidates for the j component
        J = min(length(index), W_max);  % the number of non-void subsets
        index = index(1:J) + 1; % select best subsets and account for the void measurement subset
        
        % size(local_log_w) = (curr_J * nb_meas) x nb_part
        
        curr_log_w = [local_log_w(1,:); local_log_w(index,:)];
        
        curr_paths = [local_paths(1,:); local_paths(index,:)];
        curr_J = J + 1;  % save current number of subsets
    end % for each sensor
    
    % final paths information for j_th PHD component
    % each column of curr_paths corresponds to a sensor while the lines are different combinations of the sensor measurements 
    best_paths{j} = curr_paths;
    
    x_paths{j} = [];  %  initialise the resampled particles
    
    curr_w = exp(curr_log_w);  % particle wights
    curr_r = sum(curr_w, 2)';  % weights of subsets (also serves to normalize the particle cloud of each subset)
    
    for i = 1:curr_J  % resample for each of the selected subsets
        w_paths{j}(i,:)    = curr_w(i,:) / curr_r(i); % normalize particles
        nb_eff_samp = 1/sum((w_paths{j}(i,:)).^2);    % effective sample size
        
        if nb_eff_samp <= (filter.nb_eff_part)             % resampling
            [ X_tilde, w_paths{j}(i,:), ~ ] = Resampling( curr_x, w_paths{j}(i,:) , nb_part, 'stratified' ); % resample particles
            x_paths{j}   = [x_paths{j}; X_tilde]; % store resampled particles
        else % no resampling is needed
            x_paths{j}   = [x_paths{j}; curr_x];
        end        
    end
        
    % compute final weights of subsets
    best_paths_weights{j} = curr_r;
       
end % end for each PHD component


%% generate the collection of paths - all_paths
all_paths = [];
for j = 1:J_kk1
    [all_paths] = [all_paths; best_paths{j}];
end

% Getting rid of the line [-1 -1 ... -1] and any non-unique lines
all_paths = unique(all_paths, 'rows'); % due to unique, all_paths will be sorted and component information lost
index = sum(all_paths ~= -1,2) == 0; % eliminate rows of the type [-1 -1 ... -1]
all_paths(index,:) = [];
n_paths = size(all_paths,1); % nb of linea in all_paths

end % end function find_best_subsets


function [n_partitions, unique_partitions, kappa_P] = find_best_partitions(J_kk1, trellis_paths, trellis_weights, max_partitions, clutter_rate)
% Function that constructs the best partitions from the set of subsets
% returned by the find_best_subsets function.

n_partitions = 1;
partitions_list{1} = []; % we start with a single empty partition. The list of partitions will inrease after processing the subsets of each PHD.
unique_partitions = zeros(1,J_kk1);
w_partitions = 1;

for r = 1:J_kk1 % for each predicted PHD component
    n_temp = 0; w_temp = []; temp_list = {}; temp_up = [];
    
    for i = 1:n_partitions % each preavious partition will branck into severall new partitions by appending subsets of the r-th PHD component
        % trellis_paths{r} contains the W subsets of the r-th PHD component 
        for l = 1:size(trellis_paths{r},1) 
            % check to see if an observation contained in the current
            % trellis path trellis_paths{r}(l,:) is not already contained
            % in the partition partitions_list{i}.
            if ~overlap(partitions_list{i},trellis_paths{r}(l,:)) 
                n_temp = n_temp + 1; % number of partitions
                temp_up(n_temp,:) = unique_partitions(i,:);
                
                if sum(trellis_paths{r}(l,:) == -1) ~=  length(trellis_paths{r}(l,:)) % if not the empty path
                    
                    % formation of a new partition by appending trellis_paths{r}(l,:)
                    % trellis_paths{r}(l,:) is a row of length s which contains the indices of observations (one for each of the s sensors) contained in the subset W 
                    temp_list{n_temp} = [partitions_list{i}; trellis_paths{r}(l,:)]; 
                    temp_up(n_temp,r) = l; % remember which subset was added to this partition                   
                    w_temp(n_temp)    = w_partitions(i)*trellis_weights{r}(l); % this represents the partition weight d_P = prod(d_W)
                    
                else                    
                    % trellis_path_index{r}(l) = 0 signals that
                    % trellis_paths{r}(l,:) = [-1 -1 ... -1] (ie empty path)
                    % Addig a void subset (empty path) to a partition does not change the score of that partition. 
                    % partitions_list{i} is only necessary for checking the overlap. Overlap is not occouring for the void association -1, hence trellis_paths{r}(l,:) is not appended to temp list 
                    temp_list{n_temp} = partitions_list{i};                   
                    w_temp(n_temp)    = w_partitions(i) ;
                                        
                end
            end
        end
    end
    
    % keep only unique partitions
    [temp_up,index,~] = unique(temp_up,'rows');
    temp_list = temp_list(index);
    w_temp = w_temp(index);
    n_temp = size(index,1);
    
    % keep partitions with highest weights
    n_partitions      = min(max_partitions,n_temp);
    [~, index]        = sort(w_temp,'descend');
    partitions_list   = temp_list(index(1:n_partitions)); % update the partitions list
    unique_partitions = temp_up(index(1:n_partitions),:);
    w_partitions      = w_temp(index(1:n_partitions));
    
end

% compute Kappa_P for each partition
kappa_P = zeros(size(w_partitions));
for i =1:n_partitions    
    nb_obs = sum(sum( partitions_list{i} ~= -1 ));    
    kappa_P(i) =  clutter_rate^(-nb_obs);      
end


end  % end function find_best_partitions

function does_overlap = overlap(partition, path)
% function to verify if the observation subset 'path' overlaps with the 'partition'
does_overlap = 0;

p_size = size(partition,1);
nb_sensors = size(path,2);

p = 0;
while ~does_overlap && p < p_size
    p = p + 1;
    temp = partition(p,:);
    for s = 1:nb_sensors
        if (temp(s) ~= -1) && (path(s) ~= -1) && (temp(s) == path(s))
            does_overlap = 1;
            break;
        end
    end
end

end % end function overlap
 

function cphd_out = prune_merge(cphd_in, filter, model)
% prune PHDi components with weights below a given threshold and merge
% close components. 

ind = find( cphd_in.r > filter.comp_threshold );
not_merged = ind;

merged = 0;  % number of resulting (after merger) components
ww_k = cell(1,cphd_in.J);
w_k  = cell(1,cphd_in.J);

x_mean = zeros(model.kin.x_dim, cphd_in.J);  % compute the means for all components
for i = 1:cphd_in.J
  w_k{i} = cphd_in.w{i} / sum(cphd_in.w{i});    % making sure the w_k{i} is a true pdf
  ww_k{i} = w_k{i} * cphd_in.r(i);        % the r_k weighted particle weights are used in the merging procedure
  x_mean(:,i)  = cphd_in.x{i} * w_k{i}';  % component mean
end

% Go through the components and join close (Mahalanobis distance) components
r_merge = []; w_merge = {}; x_merge = {}; 
while ~isempty(not_merged)  
    
    [~, ind_max] = max(cphd_in.r(not_merged));
    
    %ind_max = not_merged(1);
    ind_max = not_merged(ind_max);
   
    ind_rest = not_merged(~(not_merged==ind_max));

    ind_merged = ind_max;
    
    % the mean of ind_max component
    x_mean_max = x_mean(:,ind_max); 
    % the centered particles of the ind_max component
    x_max_centered = cphd_in.x{ind_max} - repmat(x_mean_max,1, filter.nb_part);
    % sample covariance estimate of the ind_max component
    P_to_invert = (x_max_centered .* repmat(cphd_in.w{ind_max}, model.kin.x_dim,1)) * x_max_centered';% + 1 * ones(model.x_dim);
        
    [U, S, V]= svd(P_to_invert);
    s = diag(S);
    k = sum(s > 1e-9);
     
    if k<=2  
        inverseP = diag(ones(1, model.kin.x_dim));  % use Euclidian distance in this case
    else
        inverseP =  (U(:, 1: k)* diag(1./ s(1: k))* V(:, 1: k)')';  % construct inverse matrix
    end

    % Compute the differences betwen the centers (i.e. means) of ind_max and ind_rest components
    m_minus = bsxfun(@minus, x_mean_max, x_mean(:,ind_rest));
    mTPm = sum(m_minus'.*(inverseP*m_minus)',2);    % dot product computes the diagonal terms
    ind_merged = horzcat(ind_max, ind_rest(mTPm < filter.merge_threshold)); 
    
    merged = merged+1;
    r_merge(merged) = sum(cphd_in.r(ind_merged));  % sum the weights of the components to be merged
 
    if length(ind_merged) > 1 % if there are >1 components to merge    
        x_2merge = [cphd_in.x{ind_merged}];
        w_2merge = [ww_k{ind_merged}];  % use the r_k weighted particle weights for resampling
        w_2merge = w_2merge / sum(w_2merge);  

        % resample particles from the joined particle clouds to be merged
        [ x_tilde, w_tilde, ~ ] = Resampling( x_2merge, w_2merge, filter.nb_part, 'stratified' ); 
        
        x_merge{merged} = x_tilde; 
        w_merge{merged} = w_tilde;         
    else % if there are no componets close to ind_max
        x_merge{merged} = cphd_in.x{ind_max};
        w_merge{merged} = cphd_in.w{ind_max};
    end  
    % do not consider the already merged components in future mergings
    not_merged = setdiff(not_merged, ind_merged);%not_merged(~ismembc(not_merged,sort(ind_merged)));
end 

% Prepare output structure
J = min(length(r_merge),filter.comp_max);
cphd_out.J = J;  % set the number of resulting PHD components
cphd_out.cdn_pmf = cphd_in.cdn_pmf; % cardinality distribution doesn't change

[~, sorted] = sort(r_merge,'descend');
cphd_out.r = r_merge(sorted(1:J));

for i = 1:J
    cphd_out.w{i} = w_merge{sorted(i)};
    cphd_out.x{i} = x_merge{sorted(i)};   
end


end % end function prune


function cphd_out = cap_tracks(cphd_in, N_cap)
% Cap total number of PHD components to a specified maximum

if cphd_in.J > N_cap
    [~, idxsort] = sort(cphd_in.r,'descend');
    idxkeep = idxsort(1:N_cap);

    cphd_out.J = length(idxkeep);
    cphd_out.r = cphd_in.r(idxkeep);
    cphd_out.cdn_pmf = cphd_in.cdn_pmf; %cardinality doesn't change
    for i = 1:length(idxkeep)
        cphd_out.x{i} = cphd_in.x{idxkeep(i)};
        cphd_out.w{i} = cphd_in.w{idxkeep(i)};
    end   
else
    cphd_out = cphd_in;
end

end % end function cap


function [X_hat, N_hat] = extract_estimates(cphd)
% State inference from the updated multi PHD distribution

% MAP cardinality estimate
[~, N_hat] = max(cphd.cdn_pmf);
N_hat = N_hat - 1;

[~, ind] = sort(cphd.r,'descend');
N_hat = min(N_hat,length(ind));   % to avoid array access overflow
ind = ind(1:N_hat);

X_hat = [];
for i = 1:N_hat
    % pick highest weight components as targets
    X_hat(:,i) = cphd.x{ind(i)} * cphd.w{ind(i)}(:) / sum( cphd.w{ind(i)}(:) );
end

end % end function extract estimates


function display_diaginfo( k, est, filter, H_predict, H_posterior, H_prune, H_cap)
% Display diagnostic information on the fly
if ~strcmp(filter.run_flag,'silence')
    disp([' time= ',num2str(k),...
          ' #est card=' num2str(est.state_n_hat(k),4),...
          ' #comp pred=' num2str(H_predict,4),...
          ' #comp post=' num2str(H_posterior,4),...
          ' #comp pruned=',num2str(H_prune),...
          ' #comp updt=',num2str(H_cap,4)   ]);
end
end % end function display