function model = gen_model_bearing_doppler( fc, pD, sensor_pos, clutter_rate)
% Function that generates basic simulation params. for the non-linear case with Bearing Doppler observation model. 
% The state vector is composed of [x; y; x_dot; y_dot] while the
% observation vector is [bearing_deg; doppler_shift_Hz].
%
% Contact: augustin.saucan@mail.mcgill.ca

% basic parameters
model.kin.x_dim = 4;   %dimension of state vector
model.obs.z_dim = 2;   %dimension of observation vector
model.kin.v_dim = 4;   %dimension of process noise
model.obs.w_dim = 2;   %dimension of observation noise


%% dynamical model parameters (CT model)
% state transformation given by gen_newstate_fn, transition matrix is N/A in non-linear case
model.kin.T = 1;                         %sampling period
model.kin.F = [eye(2,2) model.kin.T*eye(2,2); zeros(2,2) eye(2,2)]; % state transition matrix
model.kin.sigma_v = 1;
model.kin.Q = model.kin.sigma_v^2 * [(model.kin.T^3/3)*eye(2,2) (model.kin.T^2/2)*eye(2,2); (model.kin.T^2/2)*eye(2,2) (model.kin.T)*eye(2,2)];   
model.kin.sqrt_Q = sqrtm(model.kin.Q);

model.kin.prop_fn = @(x,v) gen_newstate_fn(model,x,v);                     % define the target state transition kernel, i.e., f_{k+1}(x_{k+1} | x_k ) 

% survival/death parameters
model.kin.pS = .99;                                                        % define p_S as a constant 
model.kin.compute_ps = @(x) compute_pS(model, x);                          % define p_S as a function of the target state x

% kinematical model used for simmulating the true tracks
model.kin.sim.T = 1;                         %sampling period
model.kin.sim.F = [eye(2,2) model.kin.T*eye(2,2); zeros(2,2) eye(2,2)]; % state transition matrix
model.kin.sim.sigma_v = 0.25;
model.kin.sim.Q = model.kin.sim.sigma_v^2 * [(model.kin.T^3/3)*eye(2,2) (model.kin.T^2/2)*eye(2,2); (model.kin.T^2/2)*eye(2,2) (model.kin.T)*eye(2,2)];   
model.kin.sim.sqrt_Q = sqrtm(model.kin.sim.Q);


%% birth parameters 
model.birth.xy_sym  = 400;                                      % parameter used to control birth positions
model.birth.J_gamma = 4;                                        % number of births
model.birth.r_gamma = [0.1 0.1 0.1 0.1];                        % weights of birth components
model.birth.m_gamma = [model.birth.xy_sym model.birth.xy_sym 0 0; -model.birth.xy_sym -model.birth.xy_sym 0 0; model.birth.xy_sym -model.birth.xy_sym 0 0; -model.birth.xy_sym model.birth.xy_sym 0 0]';  % birth means
model.birth.P_gamma = diag([40 40 25 25]);                      % birth covariances
model.birth.mix_w_gamma = [1 1 1 1];                            % mixture weights for birth pdf (here we only take one Gaussian per component)
model.birth.sqrt_P_gamma = sqrtm(model.birth.P_gamma);                                                                % squared root of birth covariance matrix


%% observation model prams
model.obs.sensor_pos = sensor_pos;
model.obs.nb_sensors = size(sensor_pos, 2);                                % number of sensors
model.obs.xrange = [-1000 1000];                                           % monitoring region dimensions [m]
model.obs.yrange = [-1000 1000];                                         
model.obs.celerity = 1450;                                                 % sound speed in water [m/s]
model.obs.fc       = fc;                                                   % cental frequency [Hz]
model.obs.lambda   = model.obs.celerity / model.obs.fc;                    % wavelength of the target-emitted signal [m]      
model.obs.sigma_theta   = 1;                                               % bearing std [degrees]
model.obs.sigma_doppler = 0.7;                                             % doppler shift std [Hz]
model.obs.R = diag([model.obs.sigma_theta model.obs.sigma_doppler].^2);    % measurement noise covariance

% detection parameters
model.obs.pD = pD;                                                         % define p_D as a constant
model.obs.compute_pD = @(x, s_idx) compute_pD(model, x, s_idx);            % define p_D as a function of target state and sensor index

model.obs.angle_wrap  = @(x)  [wrapTo180(x(1,:)); x(2:end,:)];
model.obs.gen_sen_obs = @(s_idx, x, w) gen_observation_fn_bearing_doppler(model, s_idx, x, w);   % define the observation function
model.obs.lik_fn      = @(Z, Hx, R)  log( mvnpdf(model.obs.angle_wrap(Z - Hx)', zeros(1, model.obs.z_dim), R ) )'; % the measurment unit for the likelihood is 1/deg/Hz.
%model.obs.lik_fn      = @(Z, Hx, R)  loggausspdf( model.obs.angle_wrap(Z - Hx), zeros(model.obs.z_dim,1), R);

% clutter parameters
model.obs.lambda_c = clutter_rate;                        %poisson average rate of uniform clutter (per scan)
model.obs.range_c  = [ -180 180; -100 100 ];              %uniform clutter on Bearing-Doppler space
model.obs.pdf_c    =  1/prod(model.obs.range_c(:,2)-model.obs.range_c(:,1)); % uniform clutter density. The measurment unit is 1/deg/Hz.


% OSPA parameters
model.ospa_cutoff = 100;
model.ospa_order  = 1;

end


