function model = gen_model_linGauss( pD )
% Function that generates basic simulation params. for the non-linear case with Bearing Doppler observation model. 
%
% Contact: augustin.saucan@mail.mcgill.ca


% basic parameters
model.kin.x_dim = 4;   %dimension of state vector
model.obs.z_dim = 2;   %dimension of observation vector
model.kin.v_dim = 4;   %dimension of process noise
model.obs.w_dim = 2;   %dimension of observation noise


%% dynamical model parameters (CT model)
% state transformation given by gen_newstate_fn, transition matrix is N/A in non-linear case
model.kin.T = 1;                         %sampling period
model.kin.F = [eye(2,2) model.kin.T*eye(2,2); zeros(2,2) eye(2,2)]; % state transition matrix
model.kin.sigma_v = 1;
model.kin.Q = model.kin.sigma_v^2 * [(model.kin.T^3/3)*eye(2,2) (model.kin.T^2/2)*eye(2,2); (model.kin.T^2/2)*eye(2,2) (model.kin.T)*eye(2,2)];   
model.kin.sqrt_Q = sqrtm(model.kin.Q);

model.kin.prop_fn = @(x,v) gen_newstate_fn(model,x,v);                     % define the target state transition kernel, i.e., f_{k+1}(x_{k+1} | x_k ) 

% survival/death parameters
model.kin.pS = .99;                                                        % define p_S as a constant 
model.kin.compute_ps = @(x) compute_pS(model, x);                          % define p_S as a function of the target state x

% kinematical model used for simmulating the true tracks
model.kin.sim.T = 1;                         %sampling period
model.kin.sim.F = [eye(2,2) model.kin.T*eye(2,2); zeros(2,2) eye(2,2)]; % state transition matrix
model.kin.sim.sigma_v = 0.25;
model.kin.sim.Q = model.kin.sim.sigma_v^2 * [(model.kin.T^3/3)*eye(2,2) (model.kin.T^2/2)*eye(2,2); (model.kin.T^2/2)*eye(2,2) (model.kin.T)*eye(2,2)];   
model.kin.sim.sqrt_Q = sqrtm(model.kin.sim.Q);


%% birth parameters 
model.birth.xy_sym  = 400;                                       % birth positions
model.birth.J_gamma = 4;                                        % number of births
model.birth.r_gamma = [0.1 0.1 0.1 0.1];                        % weights of birth components
model.birth.m_gamma = [model.birth.xy_sym model.birth.xy_sym 0 0; -model.birth.xy_sym -model.birth.xy_sym 0 0; model.birth.xy_sym -model.birth.xy_sym 0 0; -model.birth.xy_sym model.birth.xy_sym 0 0]';  % birth means
model.birth.P_gamma = diag([60 60 25 25]);                      % birth covariances
model.birth.mix_w_gamma = [1 1 1 1];                            % mixture weights for birth pdf (here we only take one Gaussian per component)
model.birth.sqrt_P_gamma = sqrtm(model.birth.P_gamma);                                                                % squared root of birth covariance matrix

%% observation model prams
model.obs.nb_sensors = size(pD, 2);                                        % number of sensors
model.obs.xrange = [-1000 1000];                                           % monitoring region dimensions in meters
model.obs.yrange = [-1000 1000];
model.obs.H = [eye(model.obs.z_dim) zeros(model.obs.z_dim,model.obs.z_dim)];   % observation matrix      
model.obs.sigma_r   = 10;                                                  % observation noise std in meters
model.obs.R = model.obs.sigma_r^2 * eye(model.obs.z_dim);                  % measurement noise covariance

% detection parameters
model.obs.pD = pD;                                                         % define p_D as a constant

model.obs.gen_sen_obs = @(s_idx, x, w)  gen_observation_fn_linGauss(model, s_idx, x, w);   % define the observation function
model.obs.lik_fn      = @(Z, Hx, R)  log( mvnpdf((Z - Hx)', zeros(1, model.obs.z_dim), R ) )';


% clutter parameters
model.obs.lambda_c = 5;                                   %poisson average rate of clutter (per scan) and identical for all sensors (this could be made sensor dependent in future releases)
model.obs.range_c  = [ -1000 1000; -1000 1000 ];          %uniform clutter on XY
model.obs.pdf_c    = 1/prod(model.obs.range_c(:,2)-model.obs.range_c(:,1)); %uniform clutter density

% Plot parameters
model.plot.xrange = [-700 830];
model.plot.yrange = [-600 630];


% OSPA parameters
model.ospa_cutoff = 100;
model.ospa_order  = 1;


end


