function truth = gen_truth(model, switch_noise, track_case, varargin)
% function truth = gen_truth(model, switch_noise, track_case, varargin)
% Function that generates a set of target trajectories and return a
% structure hodling the ground-truth tracks.
%  OUTPUT:  
%        truth.K  - number of scans (ie.e, number of samples)
%        truth.X  - cell array that holds the target state vectors for each of the truth.K samples 
%        truth.N  - number of targets in each scan
%        truth.L  - track labels (currently not employed)
%        truth.track_list - a track index used to identify the different
%                           state vectors in a cell oftruth.X (usefull for debugging)
%        truth.total_tracks - number of tracks
%  INPUT: 
%       model  - structure holding kinematic and other params.
%       switch_noise - select whether or not to have process noise
%       track_case   - select specific set of tracks (currently only 1 is defined)
%       varargin{1}  - (optional) if switch_noise == 'noise' then varargin{1} given the process noise seed 
%
% Contact: augustin.saucan@mail.mcgill.ca


if track_case == 1  % first set of tracks
    
    %variables
    truth.K = 100;                         % length of data/number of scans
    truth.X = cell(truth.K,1);             % ground truth for states of targets
    truth.N = zeros(truth.K,1);            % ground truth for number of targets
    truth.L = cell(truth.K,1);             % ground truth for labels of targets (k,i)
    truth.track_list = cell(truth.K,1);    % absolute index target identities (plotting)
    truth.total_tracks = 0;                % total number of appearing tracks
    
    %target initial states and birth/death times
    nbirths = 8;                           % total number of birthed targets
    xy_sym = model.birth.xy_sym;
    
    xstart(:,1)  = [ -xy_sym  xy_sym   5  2 ]';        tbirth(1) = 1;     tdeath(1) = 55;
    xstart(:,2)  = [ -xy_sym -xy_sym  -3 -4 ]';        tbirth(2) = 5;     tdeath(2) = 25;
    xstart(:,3)  = [ -xy_sym  xy_sym  -4 -8 ]';        tbirth(3) = 20;    tdeath(3) = 80;   
    xstart(:,4)  = [  xy_sym -xy_sym  -5  1 ]';        tbirth(4) = 35;    tdeath(4) = truth.K+1;
    xstart(:,5)  = [  xy_sym  xy_sym  -4  5 ]';        tbirth(5) = 1;     tdeath(5) = 25;
    xstart(:,6)  = [ -xy_sym -xy_sym  -4 12 ]';        tbirth(6) = 40;    tdeath(6) = truth.K+1;  
    
    % last two targets will have deterministic trajectories
    xstart(:,7)  = [  xy_sym  xy_sym   0 0 ]';        tbirth(7) = 10;    tdeath(7) = truth.K+1;
    xstart(:,8)  = [  xy_sym  -xy_sym  0 0 ]';        tbirth(8) = 10;    tdeath(8) = truth.K+1;
    
    truth.tbirth = tbirth;
    truth.xstart = xstart;
    
    seed_ini = rng;  % get current seed  (rng seed will be restored after the tracks were generated)
    
    % set up the Mersenne Twister generator for the repeatability of the simulated tracks
    if strcmp(switch_noise, 'noise') && (~isempty(varargin))
        stream = RandStream('mt19937ar','seed',varargin{1});
        RandStream.setGlobalStream(stream);
    end
    
    % generate the tracks
    for targetnum = 1:(nbirths-2)
        targetstate = xstart(:,targetnum);
        for k = tbirth(targetnum):min(tdeath(targetnum),truth.K)
            
            if strcmp(switch_noise, 'noise')
                V = mvnrnd([0,0,0,0], model.kin.sim.Q, 1)' ;
            else
                V = [0,0,0,0]';
            end
            
            targetstate = model.kin.prop_fn(targetstate, V); 
            
            truth.X{k}= [truth.X{k} targetstate];
            truth.track_list{k} = [truth.track_list{k} targetnum];
            truth.N(k) = truth.N(k) + 1;
        end
    end
    truth.total_tracks = nbirths;
    
    % Target 7 (deterministic track)
    targetstate = xstart(:,7);  %initial point
    x_line = linspace(1, 400, tdeath(7)-tbirth(7) + 1);  % create track positions as an decreasing exponenitial of x_line with rate lambda_exp
    lambda_exp = 0.005;                                  % exp. rate
    targetstate_prev = xstart(:,7);
    for k = tbirth(7):min(tdeath(7),truth.K)
        
        targetstate_xy = [(x_line(k - tbirth(7) +1 ) +400); 400*exp(-(x_line(k - tbirth(7) +1 ))*lambda_exp)];  % compute new positions (X, Y) 
        targetstate_dot_xy =  (targetstate_xy -  targetstate_prev(1:2))/model.kin.T;                            % compute velocity 
        targetstate = [targetstate_xy; targetstate_dot_xy];                                                     % foem state vector
        
        truth.X{k}= [truth.X{k} targetstate];
        truth.track_list{k} = [truth.track_list{k} targetnum];
        truth.N(k) = truth.N(k) + 1;
        
        targetstate_prev = targetstate;
    end
    
    % Target 8 (deterministic track similarly to 7)
    lambda_exp = 0.005;
    targetstate = xstart(:,8);
    x_line = linspace(1, 400, tdeath(8)-tbirth(8) + 1);
    targetstate_prev = xstart(:,8);
    for k = tbirth(8):min(tdeath(8),truth.K)
        
        
        targetstate_xy = [(x_line(k - tbirth(8) +1 ) +400); -400*exp(-(x_line(k - tbirth(8) +1 ))*lambda_exp)];
        targetstate_dot_xy =  (targetstate_xy -  targetstate_prev(1:2))/model.kin.T;
        targetstate = [targetstate_xy; targetstate_dot_xy];
        
        truth.X{k}= [truth.X{k} targetstate];
        truth.track_list{k} = [truth.track_list{k} targetnum];
        truth.N(k) = truth.N(k) + 1;
        
        targetstate_prev = targetstate;
    end
    
    rng(seed_ini);  % restore initial seed
    
else  % define a new set of tracks
    
    fprintf(2, '\n Please define a new set of tracks! \n');    
    
    
end % if track_case
end
