function [ MC_results, truth ] = run_MC_bearing_doppler(filter_params, sensor_pos, lambda_c_var, pD_var, MC)
% Function that runs multiple MC simulations for various combinations of
% number of sensors and probabilities of detection.

nb_lambda_c_val = length(lambda_c_var);

nb_sensors = size(sensor_pos, 2); % number of sensors

% We genereate ground truth tracks only once, but for this we need target
% kinematic model. Hence we first generate the model structure with any choice of observation parameters.  
model = gen_model_bearing_doppler(300, pD_var(1) * ones(1,nb_sensors), sensor_pos, lambda_c_var(1));            % set carrier freq., pD, sensor positions and clutter rate
% Second, using the kinematic model we create tracks.
truth = gen_truth(model,'noise', 1, 50);


for lambda_c_index = 1:nb_lambda_c_val

    for  pD_index = 1:length(pD_var)
        
        % regenerate model for specified simulation parameters
        pD = pD_var(pD_index) * ones(1,nb_sensors);                        % set prob. of detection for sensors           
        model = gen_model_bearing_doppler(300, pD, sensor_pos, lambda_c_var(lambda_c_index));            % set carrier freq., pD, sensor positions and clutter rate    

        fprintf('\n ######################################################################\n');
        fprintf('\n \t \t \t '); s = datestr(now); disp(s);                                                       % display current timestamp        
        fprintf('\n %d MC runs for lambda_c = %d and pD = [%s]\n', MC, lambda_c_var(lambda_c_index), sprintf(' %.1f ', pD));  % display in simmulation params
        fprintf('\n ######################################################################\n');
        
        % run MC simmulation for given model and tracks
        [ local_output ] = func_MC_multi_sensor(filter_params, model, truth, MC);
                
        % save MC simulation results 
        MC_results(lambda_c_index, pD_index).ic_cbmember_ukf  = local_output.ic_cbmember_ukf;
        MC_results(lambda_c_index, pD_index).ic_cbmember_smc  = local_output.ic_cbmember_smc;
        
        MC_results(lambda_c_index, pD_index).ms_member_ukf  = local_output.ms_member_ukf;
        MC_results(lambda_c_index, pD_index).ms_member_smc  = local_output.ms_member_smc;

        MC_results(lambda_c_index, pD_index).ms_cphd_ukf    = local_output.ms_cphd_ukf;
        MC_results(lambda_c_index, pD_index).ms_cphd_smc    = local_output.ms_cphd_smc;
        
        MC_results(lambda_c_index, pD_index).ms_tcphd_ukf   = local_output.ms_tcphd_ukf;
        MC_results(lambda_c_index, pD_index).ms_tcphd_smc   = local_output.ms_tcphd_smc;
        
        % also save model params
        MC_results(lambda_c_index, pD_index).model       = model;
       
    end
    

end
end % end function

function [mc_output] = func_MC_multi_sensor(filter_params, model, truth, MC)


ms_member_smc_params = filter_params.ms_member_smc;
mc_member_smc_ospa = zeros(MC,truth.K);
mc_member_smc_x_est   = cell(1,MC);
mc_member_smc_card   = zeros(MC,truth.K);
mc_member_smc_all_time = zeros(MC,1);
mc_member_smc_rkk     = cell(MC);

ms_member_ukf_params = filter_params.ms_member_ukf;
mc_member_ukf_ospa = zeros(MC,truth.K);
mc_member_ukf_x_est   = cell(1,MC);
mc_member_ukf_card   = zeros(MC,truth.K);
mc_member_ukf_all_time = zeros(MC,1);
mc_member_ukf_rkk     = cell(MC);

ms_cphd_smc_params = filter_params.ms_cphd_smc;
mc_cphd_smc_ospa = zeros(MC,truth.K);
mc_cphd_smc_x_est   = cell(1,MC);
mc_cphd_smc_card   = zeros(MC,truth.K);
mc_cphd_smc_all_time = zeros(MC,1);
mc_cphd_smc_rkk     = cell(MC);

ms_cphd_ukf_params = filter_params.ms_cphd_ukf;
mc_cphd_ukf_ospa = zeros(MC,truth.K);
mc_cphd_ukf_x_est   = cell(1,MC);
mc_cphd_ukf_card   = zeros(MC,truth.K);
mc_cphd_ukf_all_time = zeros(MC,1);
mc_cphd_ukf_rkk     = cell(MC);

ms_tcphd_smc_params = filter_params.ms_tcphd_smc;
mc_tcphd_smc_ospa = zeros(MC,truth.K);
mc_tcphd_smc_x_est   = cell(1,MC);
mc_tcphd_smc_card   = zeros(MC,truth.K);
mc_tcphd_smc_all_time = zeros(MC,1);
mc_tcphd_smc_rkk     = cell(MC);

ms_tcphd_ukf_params = filter_params.ms_tcphd_ukf;
mc_tcphd_ukf_ospa = zeros(MC,truth.K);
mc_tcphd_ukf_x_est   = cell(1,MC);
mc_tcphd_ukf_card   = zeros(MC,truth.K);
mc_tcphd_ukf_all_time = zeros(MC,1);
mc_tcphd_ukf_rkk     = cell(MC);

ms_ic_cbmember_smc_params =  filter_params.ic_cbmember_smc;
mc_ic_cbmember_smc_ospa   = zeros(MC,truth.K);
mc_ic_cbmember_smc_x_est   = cell(1,MC);
mc_ic_cbmember_smc_card   = zeros(MC,truth.K);
mc_ic_cbmember_smc_all_time = zeros(MC,1);
mc_ic_cbmember_smc_rkk     = cell(MC);

ms_ic_cbmember_ukf_params =  filter_params.ic_cbmember_ukf;
mc_ic_cbmember_ukf_ospa   = zeros(MC,truth.K);
mc_ic_cbmember_ukf_x_est   = cell(1,MC);
mc_ic_cbmember_ukf_card   = zeros(MC,truth.K);
mc_ic_cbmember_ukf_all_time = zeros(MC,1);
mc_ic_cbmember_ukf_rkk     = cell(MC);


% set up the increments of the MC fprintf
if MC > 500
    notch = 100;
elseif MC > 100
    notch = 50;
elseif MC >= 30
    notch = 10;
else
    notch = 1;
end

for mc = 1:MC
    
    if (mod(mc,notch) == 0) || (mc == 1)  % print run number only for multiples of 100
        fprintf('\n    -- running for mc = %d ', mc)
    end
          
    meas  =  gen_meas_bearing_doppler(model, truth, []);
    
    % Single run results for SMC MS-MeMBer 
    est_ms_member_smc = filter_ms_member_smc(ms_member_smc_params, model, meas,truth);                                                % run filter                                                                               % compute time
     
    mc_member_smc_ospa(mc,:)       =  est_ms_member_smc.ospa;
    mc_member_smc_x_est{mc}        =  est_ms_member_smc.state_X_hat;    
    mc_member_smc_card(mc,:)       =  est_ms_member_smc.state_n_hat;
    mc_member_smc_all_time(mc)     =  est_ms_member_smc.all_time;
    mc_member_smc_rkk{mc}          =  est_ms_member_smc.rkk;

    % Single run results for UKF MS-MeMBer
    est_ms_member_ukf = filter_ms_member_ukf(ms_member_ukf_params, model, meas, truth);                                                % run filter                                                                               % compute time
     
    mc_member_ukf_ospa(mc,:)       =  est_ms_member_ukf.ospa;
    mc_member_ukf_x_est{mc}        =  est_ms_member_ukf.state_X_hat;    
    mc_member_ukf_card(mc,:)       =  est_ms_member_ukf.state_n_hat;
    mc_member_ukf_all_time(mc)     =  est_ms_member_ukf.all_time; 
    mc_member_ukf_rkk{mc}          =  est_ms_member_ukf.rkk;

    % Single run results for SMC MS-CPHD
    est_ms_cphd_smc = filter_ms_cphd_smc(ms_cphd_smc_params, model, meas, truth);                                                % run filter                                                                            % compute time
     
    mc_cphd_smc_ospa(mc,:)       =  est_ms_cphd_smc.ospa;
    mc_cphd_smc_x_est{mc}        =  est_ms_cphd_smc.state_X_hat;    
    mc_cphd_smc_card(mc,:)       =  est_ms_cphd_smc.state_n_hat;
    mc_cphd_smc_all_time(mc)     =  est_ms_cphd_smc.all_time;    
    mc_cphd_smc_rkk{mc}          =  est_ms_cphd_smc.rkk;

    % Single run results for UKF MS-CPHD
    est_ms_cphd_ukf = filter_ms_cphd_ukf(ms_cphd_ukf_params, model, meas, truth);                                                % run filter

    mc_cphd_ukf_ospa(mc,:)       =  est_ms_cphd_ukf.ospa;
    mc_cphd_ukf_x_est{mc}        =  est_ms_cphd_ukf.state_X_hat;
    mc_cphd_ukf_card(mc,:)       =  est_ms_cphd_ukf.state_n_hat;
    mc_cphd_ukf_all_time(mc)     =  est_ms_cphd_ukf.all_time;
    mc_cphd_ukf_rkk{mc}          =  est_ms_cphd_ukf.rkk;

    % Single run results for SMC MS-TCPHD 
    est_ms_tcphd_smc =  filter_ms_tcphd_smc(ms_tcphd_smc_params, model, meas, truth);                                               % run filter
     
    mc_tcphd_smc_ospa(mc,:)       =  est_ms_tcphd_smc.ospa;
    mc_tcphd_smc_x_est{mc}        =  est_ms_tcphd_smc.state_X_hat;    
    mc_tcphd_smc_card(mc,:)       =  est_ms_tcphd_smc.state_n_hat;
    mc_tcphd_smc_all_time(mc)     =  est_ms_tcphd_smc.all_time;
    mc_tcphd_smc_rkk{mc}          =  est_ms_tcphd_smc.rkk;

    % Single run results for UKF MS-TCPHD 
    est_ms_tcphd_ukf = filter_ms_tcphd_ukf(ms_tcphd_ukf_params, model, meas, truth);                                                % run filter
     
    mc_tcphd_ukf_ospa(mc,:)       =  est_ms_tcphd_ukf.ospa;
    mc_tcphd_ukf_x_est{mc}        =  est_ms_tcphd_ukf.state_X_hat;    
    mc_tcphd_ukf_card(mc,:)       =  est_ms_tcphd_ukf.state_n_hat;
    mc_tcphd_ukf_all_time(mc)     =  est_ms_tcphd_ukf.all_time; 
    mc_tcphd_ukf_rkk{mc}          =  est_ms_tcphd_ukf.rkk;

    % Single run results for SMC IC-CBMeMBer 
    est_ic_cbmember_smc = filter_ic_cbmember_smc(ms_ic_cbmember_smc_params, model, meas, truth);                                                % run filter
    
    mc_ic_cbmember_smc_ospa(mc,:)       =  est_ic_cbmember_smc.ospa;
    mc_ic_cbmember_smc_x_est{mc}        =  est_ic_cbmember_smc.state_X_hat;    
    mc_ic_cbmember_smc_card(mc,:)       =  est_ic_cbmember_smc.state_n_hat;
    mc_ic_cbmember_smc_all_time(mc)     =  est_ic_cbmember_smc.all_time; 
    mc_ic_cbmember_smc_rkk{mc}          =  est_ic_cbmember_smc.rkk;
    
   % Single run results for UKF IC-CBMeMBer 
    est_ic_cbmember_ukf = filter_ic_cbmember_ukf(ms_ic_cbmember_ukf_params, model, meas, truth);                                                % run filter
    
    mc_ic_cbmember_ukf_ospa(mc,:)       =  est_ic_cbmember_ukf.ospa;
    mc_ic_cbmember_ukf_x_est{mc}        =  est_ic_cbmember_ukf.state_X_hat;    
    mc_ic_cbmember_ukf_card(mc,:)       =  est_ic_cbmember_ukf.state_n_hat;
    mc_ic_cbmember_ukf_all_time(mc)     =  est_ic_cbmember_ukf.all_time; 
    mc_ic_cbmember_ukf_rkk{mc}          =  est_ic_cbmember_ukf.rkk;
end

% save MC runs results for SMC MS-MeMBer
mc_output.ms_member_smc.ospa = mc_member_smc_ospa;
mc_output.ms_member_smc.card = mc_member_smc_card;
mc_output.ms_member_smc.comp_time = nanmean(mc_member_smc_all_time);
mc_output.ms_member_smc.all_time = mc_member_smc_all_time;
mc_output.ms_member_smc.x_est    = mc_member_smc_x_est;
mc_output.ms_member_smc.rkk      = mc_member_smc_rkk;

% save MC runs results for UKF MS-MeMBer
mc_output.ms_member_ukf.ospa = mc_member_ukf_ospa;
mc_output.ms_member_ukf.card = mc_member_ukf_card;
mc_output.ms_member_ukf.comp_time = nanmean(mc_member_ukf_all_time);
mc_output.ms_member_ukf.all_time = mc_member_ukf_all_time;
mc_output.ms_member_ukf.x_est    = mc_member_ukf_x_est;
mc_output.ms_member_ukf.rkk      = mc_member_ukf_rkk;

% save MC runs results for SMC MS-CPHD
mc_output.ms_cphd_smc.ospa = mc_cphd_smc_ospa;
mc_output.ms_cphd_smc.card = mc_cphd_smc_card;
mc_output.ms_cphd_smc.comp_time = nanmean(mc_cphd_smc_all_time);
mc_output.ms_cphd_smc.all_time = mc_cphd_smc_all_time;
mc_output.ms_cphd_smc.x_est = mc_cphd_smc_x_est;
mc_output.ms_cphd_smc.rkk      = mc_cphd_smc_rkk;

% save MC runs results for UKF MS-CPHD
mc_output.ms_cphd_ukf.ospa = mc_cphd_ukf_ospa;
mc_output.ms_cphd_ukf.card = mc_cphd_ukf_card;
mc_output.ms_cphd_ukf.comp_time = nanmean(mc_cphd_ukf_all_time);
mc_output.ms_cphd_ukf.all_time = mc_cphd_ukf_all_time;
mc_output.ms_cphd_ukf.x_est = mc_cphd_ukf_x_est;
mc_output.ms_cphd_ukf.rkk      = mc_cphd_ukf_rkk;

% save MC runs results for SMC MS-TCPHD
mc_output.ms_tcphd_smc.ospa = mc_tcphd_smc_ospa;
mc_output.ms_tcphd_smc.card = mc_tcphd_smc_card;
mc_output.vtcphd_smc.comp_time = nanmean(mc_tcphd_smc_all_time);
mc_output.ms_tcphd_smc.all_time = mc_tcphd_smc_all_time;
mc_output.ms_tcphd_smc.x_est = mc_tcphd_smc_x_est;
mc_output.ms_tcphd_smc.rkk      = mc_tcphd_smc_rkk;

% save MC runs results for UKF MS-TCPHD
mc_output.ms_tcphd_ukf.ospa = mc_tcphd_ukf_ospa;
mc_output.ms_tcphd_ukf.card = mc_tcphd_ukf_card;
mc_output.ms_tcphd_ukf.comp_time = nanmean(mc_tcphd_ukf_all_time);
mc_output.ms_tcphd_ukf.all_time = mc_tcphd_ukf_all_time;
mc_output.ms_tcphd_ukf.x_est = mc_tcphd_ukf_x_est;
mc_output.ms_tcphd_ukf.rkk      = mc_tcphd_ukf_rkk;


% save MC runs results for SMC IC-CBMeMBer
mc_output.ic_cbmember_smc.ospa      = mc_ic_cbmember_smc_ospa;
mc_output.ic_cbmember_smc.card      = mc_ic_cbmember_smc_card;
mc_output.ic_cbmember_smc.comp_time = nanmean(mc_ic_cbmember_smc_all_time);
mc_output.ic_cbmember_smc.all_time  = mc_ic_cbmember_smc_all_time;
mc_output.ic_cbmember_smc.x_est     = mc_ic_cbmember_smc_x_est;
mc_output.ic_cbmember_smc.rkk       = mc_ic_cbmember_smc_rkk;


% save MC runs results for UKF IC-CBMeMBer
mc_output.ic_cbmember_ukf.ospa      = mc_ic_cbmember_ukf_ospa;
mc_output.ic_cbmember_ukf.card      = mc_ic_cbmember_ukf_card;
mc_output.ic_cbmember_ukf.comp_time = nanmean(mc_ic_cbmember_ukf_all_time);
mc_output.ic_cbmember_ukf.all_time  = mc_ic_cbmember_ukf_all_time;
mc_output.ic_cbmember_ukf.x_est     = mc_ic_cbmember_ukf_x_est;
mc_output.ic_cbmember_ukf.rkk       = mc_ic_cbmember_ukf_rkk;

end



