function [ MC_results, truth ] = run_MC_lin_Gauss(filter_params, sensor_param, pD_param, MC)
% Function that runs multiple MC simulations for various combinations of
% number of sensors and probabilities of detection.

nb_sensors_max = length(sensor_param);
pD_max = length(pD_param);

% We genereate ground truth tracks only once, but for this we need target
% kinematic model. Hence we first generate the model structure with any choice of observation parameters.  
model = gen_model_linGauss(pD_param(1) * ones(1,sensor_param(1)));
% generate true tracks
truth = gen_truth(model, 'noise', 1, 50 );

for s_index = 1:nb_sensors_max
    
    for  pD_index = 1:pD_max
        
        % generate model
        pD = pD_param(pD_index) * ones(1,sensor_param(s_index));             
        model = gen_model_linGauss(pD); 
        
        fprintf('\n ######################################################################\n');
        fprintf('\n \t \t \t '); s = datestr(now); disp(s);                                                       % display current timestamp
        fprintf('\n %d MC runs for for nb_sensors = %d and pD = [%s]\n',MC, sensor_param(s_index), sprintf(' %.1f ', pD));  % display in vector vector 
        fprintf('\n ######################################################################\n');
        
        % run MC simmulation for given model and tracks
        [ local_output] = func_MC_multi_sensor(filter_params, model, truth, MC);
                
        % save MC simulation results 
        MC_results(s_index, pD_index).ic_cbmember  = local_output.ic_cbmember;
        
        MC_results(s_index, pD_index).ms_member  = local_output.ms_member;

        MC_results(s_index, pD_index).ms_cphd    = local_output.ms_cphd;

        MC_results(s_index, pD_index).ms_tcphd   = local_output.ms_tcphd;
        
        % also save model params
        MC_results(s_index, pD_index).model       = model;
       
    end
    
end

end


function [mc_output] = func_MC_multi_sensor(filter_params, model, truth, MC)

ms_ic_cbmember_params = filter_params.ic_cbmember_kf;
mc_ic_cbmember_ospa = zeros(MC,truth.K);
mc_ic_cbmember_x_est   = cell(1,MC);
mc_ic_cbmember_rkk   = cell(1,truth.K);
mc_ic_cbmember_card   = zeros(MC,truth.K);
mc_ic_cbmember_all_time = zeros(MC,1);

ms_member_params = filter_params.ms_member_kf;
mc_member_ospa = zeros(MC,truth.K);
mc_member_x_est   = cell(1,MC);
mc_member_rkk   = cell(1,truth.K);
mc_member_card   = zeros(MC,truth.K);
mc_member_all_time = zeros(MC,1);

ms_cphd_params = filter_params.ms_cphd_kf;
mc_cphd_ospa = zeros(MC,truth.K);
mc_cphd_x_est   = cell(1,MC);
mc_cphd_rkk   = cell(1,truth.K);
mc_cphd_card   = zeros(MC,truth.K);
mc_cphd_all_time = zeros(MC,1);

ms_tcphd_params = filter_params.ms_tcphd_kf;
mc_tcphd_ospa = zeros(MC,truth.K);
mc_tcphd_x_est   = cell(1,MC);
mc_tcphd_rkk   = cell(1,truth.K);
mc_tcphd_card   = zeros(MC,truth.K);
mc_tcphd_all_time = zeros(MC,1);


% set up the increments of the MC fprintf
if MC > 500
    notch = 100;
elseif MC > 100
    notch = 50;
elseif MC >= 30
    notch = 10;
else
    notch = 1;
end

for mc = 1:MC
    
    if (mod(mc,notch) == 0) || (mc == 1)  % print run number only for multiples of 100
        fprintf('\n    -- running for mc = %d ', mc)
    end
          
    % Generate Measurements
    meas  =  gen_meas_linGauss(model, truth, []);

    % Single run results for MS-IC-CBMeMBer
    est_ms_ic_cbmember = filter_ic_cbmember_kf(ms_ic_cbmember_params, model, meas, truth);                                                % run filter
    
    mc_ic_cbmember_ospa(mc,:)       =  est_ms_ic_cbmember.ospa;
    mc_ic_cbmember_x_est{mc}        =  est_ms_ic_cbmember.state_X_hat;
    mc_ic_cbmember_card(mc,:)       =  est_ms_ic_cbmember.state_n_hat;
    mc_ic_cbmember_all_time(mc)     =  est_ms_ic_cbmember.all_time;
    
    % Single run results for MS-MeMBer 
    est_ms_member = filter_ms_member_kf(ms_member_params, model, meas,truth);                                                % run filter
  
    mc_member_ospa(mc,:)       =  est_ms_member.ospa;
    mc_member_x_est{mc}        =  est_ms_member.state_X_hat;    
    mc_member_card(mc,:)       =  est_ms_member.state_n_hat;
    mc_member_all_time(mc)     =  est_ms_member.all_time;
   
    % Single run results for MS-CPHD 
    est_ms_cphd = filter_ms_cphd_kf(ms_cphd_params, model, meas, truth);                                                % run filter
                                                                            
    mc_cphd_ospa(mc,:)       =  est_ms_cphd.ospa;
    mc_cphd_x_est{mc}        =  est_ms_cphd.state_X_hat;    
    mc_cphd_card(mc,:)       =  est_ms_cphd.state_n_hat;
    mc_cphd_all_time(mc)     =  est_ms_cphd.all_time;   
    
    % Single run results for MS-TCPHD 
    est_ms_tcphd = filter_ms_tcphd_kf(ms_tcphd_params, model, meas, truth);                                               % run filter                                                                            % compute time
    
    mc_tcphd_ospa(mc,:)      =  est_ms_tcphd.ospa;
    mc_tcphd_x_est{mc}       =  est_ms_tcphd.state_X_hat;    
    mc_tcphd_card(mc,:)      =  est_ms_tcphd.state_n_hat;
    mc_tcphd_all_time(mc)    =  est_ms_tcphd.all_time;
end

% save MC runs results for MS-MeMBer
mc_output.ic_cbmember.ospa     = mc_ic_cbmember_ospa;
mc_output.ic_cbmember.card     = mc_ic_cbmember_card;
mc_output.ic_cbmember.all_time = mc_ic_cbmember_all_time;
mc_output.ic_cbmember.x_est    = mc_ic_cbmember_x_est;
mc_output.ic_cbmember.rkk      = mc_ic_cbmember_rkk;

% save MC runs results for MS-MeMBer
mc_output.ms_member.ospa     = mc_member_ospa;
mc_output.ms_member.card     = mc_member_card;
mc_output.ms_member.all_time = mc_member_all_time;
mc_output.ms_member.x_est    = mc_member_x_est;
mc_output.ms_member.rkk      = mc_member_rkk;

% save MC runs results for MS-CPHD
mc_output.ms_cphd.ospa     = mc_cphd_ospa;
mc_output.ms_cphd.card     = mc_cphd_card;
mc_output.ms_cphd.all_time = mc_cphd_all_time;
mc_output.ms_cphd.x_est    = mc_cphd_x_est;
mc_output.ms_cphd.rkk      = mc_cphd_rkk;

% save MC runs results for MS-TCPHD
mc_output.ms_tcphd.ospa      = mc_tcphd_ospa;
mc_output.ms_tcphd.card      = mc_tcphd_card;
mc_output.ms_tcphd.all_time  = mc_tcphd_all_time;
mc_output.ms_tcphd.x_est     = mc_tcphd_x_est;
mc_output.ms_tcphd.rkk       = mc_tcphd_rkk;

end



