% This is a demo script that performs a MC analysis for a non-linear 
% scenario (Beraring Doppler) of the the MS-MeMBer [1], MS-CPHD [2] and 
% MS-TCPHD [1] filters.
%
% [1] A. Saucan, M. Coates, and M. Rabbat,  "A multi-sensor multi-Bernoulli 
% filter"  2016, arXiv:1609.05108 [stat.ME]
%
% [2] S. Nannuru, S. Blouin, M. Coates, and M. Rabbat,  "Multisensor CPHD 
% filter", IEEE Trans. Aerosp. and Electron. Syst., vol. 52, no. 4, pp. 
% 1834-1854, Aug. 2016, arXiv:1504.06342v2 [stat.ME]
%
% [3] B.-T. Vo, B.-N. Vo, and A. Cantoni, “The cardinality balanced 
% multi-target multi-Bernoulli filter and its implementations,”
% IEEE Trans. Signal Process., vol. 57, no. 2, pp. 409 – 423, Feb. 2009.
%
% Contact: augustin.saucan@mail.mcgill.ca

clear all
close all
clc

% restoredefaultpath     % clear previous path
addpath(genpath(pwd));   % append all folders to current path
%% Generate filtering parameters
filter_params = gen_filters();

%% specify prob. of detection and clutter rate values for which an MC simulation will be performed
pD_var = [0.3 0.5 0.9]; % prob. of detection (all sensors have identical pD) 
lambda_c_var = [5 20 40 ]; % clutter rates 
MC = 100;   % number of Monte Carlo runs

sensor_pos = [ 0      0   0  -350 350 ;  % sensor positions X coord
              350  -350   0     0   0 ];  % sensor positions Y coord
 
start = tic;   

[ MC_results, truth ] = run_MC_bearing_doppler(filter_params, sensor_pos, lambda_c_var, pD_var, MC);
total_time = toc(start);

fprintf('\n ##### I reached the end and it took %d seconds (%d hours)! #######\n', total_time , total_time/3600)

% Recompute OSPA errors if necessary 
% [ MC_results2 ] = compute_ospa_nonLin( MC_results, truth, MC );

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%% Tables %%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

lambda_index = 1;   % 1 for 10; 2 for 20; 3 for 40
pd_index = 1;    % 1 for 0.3 ; 2 for 0.5; 3 for 0.9; 

ic_cbmember_smc_avr = mean(MC_results(lambda_index, pd_index).ic_cbmember_smc.ospa, 2);
ic_cbmember_ukf_avr = mean(MC_results(lambda_index, pd_index).ic_cbmember_ukf.ospa, 2);

member_smc_avr = mean(MC_results(lambda_index, pd_index).ms_member_smc.ospa, 2);
member_ukf_avr = mean(MC_results(lambda_index, pd_index).ms_member_ukf.ospa, 2);

cphd_smc_avr = mean(MC_results(lambda_index, pd_index).ms_cphd_smc.ospa, 2);
cphd_ukf_avr = mean(MC_results(lambda_index, pd_index).ms_cphd_ukf.ospa, 2);

tcphd_smc_avr = mean(MC_results(lambda_index, pd_index).ms_tcphd_smc.ospa, 2);
tcphd_ukf_avr = mean(MC_results(lambda_index, pd_index).ms_tcphd_ukf.ospa, 2);

fprintf('\n IC_CBMeMBer SMC:')
strcat(num2str(median(ic_cbmember_smc_avr)), ' [', num2str(quantile(ic_cbmember_smc_avr,[0.25 0.75])),']', num2str(mean(MC_results(lambda_index, pd_index).ic_cbmember_smc.all_time)/truth.K), 'sec')
fprintf('\n IC_CBMeMBer UKF:')
strcat(num2str(median(ic_cbmember_ukf_avr)), ' [', num2str(quantile(ic_cbmember_ukf_avr,[0.25 0.75])),']', num2str(mean(MC_results(lambda_index, pd_index).ic_cbmember_ukf.all_time)/truth.K), 'sec')

fprintf('\n MS_MeMBer SMC:')
strcat(num2str(median(member_smc_avr)), ' [', num2str(quantile(member_smc_avr,[0.25 0.75])),']', num2str(mean(MC_results(lambda_index, pd_index).ms_member_smc.all_time)/truth.K), 'sec')
fprintf('\n MS_MeMBer UKF:')
strcat(num2str(median(member_ukf_avr)), ' [', num2str(quantile(member_ukf_avr,[0.25 0.75])),']', num2str(mean(MC_results(lambda_index, pd_index).ms_member_ukf.all_time)/truth.K), 'sec')


fprintf('\n MS_CPHD SMC:')
strcat(num2str(median(cphd_smc_avr)), ' [', num2str(quantile(cphd_smc_avr,[0.25 0.75])),']', num2str(mean(MC_results(lambda_index, pd_index).ms_cphd_smc.all_time)/truth.K), 'sec')
fprintf('\n MS_CPHD UKF:')
strcat(num2str(median(cphd_ukf_avr)), ' [', num2str(quantile(cphd_ukf_avr,[0.25 0.75])),']', num2str(mean(MC_results(lambda_index, pd_index).ms_cphd_ukf.all_time)/truth.K), 'sec')

fprintf('\n MS_TCPHD SMC:')
strcat(num2str(median(tcphd_smc_avr)), ' [', num2str(quantile(tcphd_smc_avr,[0.25 0.75])),']', num2str(mean(MC_results(lambda_index, pd_index).ms_tcphd_smc.all_time)/truth.K), 'sec')
fprintf('\n MS_TCPHD UKF:')
strcat(num2str(median(tcphd_ukf_avr)), ' [', num2str(quantile(tcphd_ukf_avr,[0.25 0.75])),']', num2str(mean(MC_results(lambda_index, pd_index).ms_tcphd_ukf.all_time)/truth.K), 'sec')

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%% Plot functions %%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Cardinality plot
[ fHand ] = plot_cardinality_bearing_doppler( MC_results(1), truth );

[ fHand ] = plot_cardinality_hist_bearing_doppler( MC_results(1), truth );


%%OSPA Box plot
[ hfig ] = box_plot_time_averaged_OSPA_pD_bearing_doppler( MC_results, pD_var, MC );

[ hfig ] = box_plot_time_averaged_OSPA_clutter_bearing_doppler( MC_results, lambda_c_var, MC );


%% Comp time Box-plot
[ hfig ] = box_plot_run_time_pD_bearing_doppler( MC_results, truth, pD_var, MC );



